#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
import warnings
from typing import Any
from typing import Dict
from typing import Optional

import hydra
import lightning as L
import torch.utils.data
from omegaconf import DictConfig

from disent.dataset import DisentDataset
from disent.dataset.transform import DisentDatasetTransform

log = logging.getLogger(__name__)


# ========================================================================= #
# DISENT DATASET MODULE                                                     #
# TODO: possible implementation outline for disent                          #
# ========================================================================= #


# class DisentDatasetModule(L.LightningDataModule):
#
#     def prepare_data(self, *args, **kwargs):
#         raise NotImplementedError
#
#     def setup(self, stage: Optional[str] = None):
#         raise NotImplementedError
#
#     # DATASET HANDLING
#
#     @property
#     def dataset_train(self) -> DisentDataset:
#         # this property should check `framework_applies_augment` to return the
#         # dataset with the correct transforms/augments applied.
#         # - train_dataloader() should create the DataLoader from this dataset object
#         raise NotImplementedError
#
#     # FRAMEWORK AUGMENTATION HANDLING
#
#     @property
#     def framework_applies_augment(self) -> bool:
#         # if we augment the data in the framework rather, we can augment on the GPU instead
#         # framework needs manual handling of this augmentation mode
#         raise NotImplementedError
#
#     def framework_augment(self, batch):
#         # the augment to be applied if `framework_applies_augment` is `True`, otherwise
#         # this method should do nothing!
#         raise NotImplementedError


# ========================================================================= #
# DATASET                                                                   #
# ========================================================================= #


class HydraDataModule(L.LightningDataModule):
    def __init__(
        self,
        data: Dict[str, Any],  # = dataset.data
        sampler: Dict[str, Any],  # = sampling._sampler_.sampler_cls
        transform: Optional[Dict[str, Any]] = None,  # = dataset.transform
        augment: Optional[Dict[str, Any]] = None,  # = augment.augment_cls
        dataloader_kwargs: Optional[Dict[str, Any]] = None,  # = dataloader
        augment_on_gpu: bool = False,  # = dsettings.dataset.gpu_augment
        using_cuda: Optional[bool] = False,  # = self.hparams.dsettings.trainer.cuda
        prepare_data_per_node: bool = True,  # DataHooks.prepare_data_per_node
        return_indices: bool = False,  # = framework.meta.requires_indices
        return_factors: bool = False,  # = framework.meta.requires_factors
    ):
        super().__init__()
        # OVERRIDE:
        self.prepare_data_per_node = prepare_data_per_node
        # save hparams
        self.save_hyperparameters()
        # ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #
        # transform: prepares data from datasets
        self.data_transform = hydra.utils.instantiate(transform)
        assert (self.data_transform is None) or callable(self.data_transform)
        # input_transform_aug: augment data for inputs, then apply input_transform
        self.input_transform = hydra.utils.instantiate(augment)
        assert (self.input_transform is None) or callable(
            self.input_transform
        )  # should be: `Callable[[torch.Tensor], torch.Tensor]`
        # ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #
        # batch_augment: augments transformed data for inputs, should be applied across a batch
        # which version of the dataset we need to use if GPU augmentation is enabled or not.
        # - corresponds to below in train_dataloader()
        if augment_on_gpu:
            self._gpu_batch_augment = DisentDatasetTransform(transform=self.input_transform)
            warnings.warn("`augment_on_gpu=True` is outdated and may no longer be equivalent to `augment_on_gpu=False`")
        else:
            self._gpu_batch_augment = None
        # ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #
        # datasets initialised in setup()
        self.dataset_train_noaug: DisentDataset = None
        self.dataset_train_aug: DisentDataset = None

    @property
    def gpu_batch_augment(self) -> Optional[DisentDatasetTransform]:
        return self._gpu_batch_augment

    def prepare_data(self) -> None:
        # *NB* Do not set model parameters here.
        # - Instantiate data once to download and prepare if needed.
        # - trainer.prepare_data_per_node affects this functions behavior per node.
        data = dict(self.hparams.data)
        if "in_memory" in data:
            del data["in_memory"]
        # create the data
        # - we instantiate the data twice, once here and once in setup otherwise
        #   things could go wrong. We try be efficient about it by removing the
        #   in_memory argument if it exists.
        log.info(f"Data - Preparation & Downloading")
        hydra.utils.instantiate(data)

    def setup(self, stage=None) -> None:
        # ground truth data
        log.info(f"Data - Instance")
        data = hydra.utils.instantiate(self.hparams.data)
        # Wrap the data for the framework some datasets need triplets, pairs, etc.
        # Augmentation is done inside the frameworks so that it can be done on the GPU, otherwise things are very slow.
        self.dataset_train_noaug = DisentDataset(
            data,
            hydra.utils.instantiate(self.hparams.sampler),
            transform=self.data_transform,
            augment=None,
            return_indices=self.hparams.return_indices,
            return_factors=self.hparams.return_factors,
        )
        self.dataset_train_aug = DisentDataset(
            data,
            hydra.utils.instantiate(self.hparams.sampler),
            transform=self.data_transform,
            augment=self.input_transform,
            return_indices=self.hparams.return_indices,
            return_factors=self.hparams.return_factors,
        )

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Training Dataset:
    #     The sample of data used to fit the model.
    # Validation Dataset:
    #     Data used to provide an unbiased evaluation of a model fit on the
    #     training dataset while tuning model hyperparameters. The
    #     evaluation becomes more biased as skill on the validation
    #     dataset is incorporated into the model configuration.
    # Test Dataset:
    #     The sample of data used to provide an unbiased evaluation of a
    #     final model fit on the training dataset.
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def train_dataloader(self):
        """
        Training Dataset: Sample of data used to fit the model.
        """
        # Select which version of the dataset we need to use if GPU augmentation is enabled or not.
        # - corresponds to above in __init__()
        if self.hparams.augment_on_gpu:
            dataset = self.dataset_train_noaug
        else:
            dataset = self.dataset_train_aug
        # ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #
        # get default kwargs
        default_kwargs = {
            "shuffle": True,
            # This should usually be TRUE if cuda is enabled.
            # About 20% faster with the xysquares dataset, RTX 2060 Rev. A, and Intel i7-3930K
            "pin_memory": self.hparams.using_cuda,
        }
        # get config kwargs
        kwargs = self.hparams.dataloader_kwargs
        if not kwargs:
            kwargs = {}
        # check required keys
        if ("batch_size" not in kwargs) or ("num_workers" not in kwargs):
            raise KeyError(
                f'`dataset.dataloader` must contain keys: ["batch_size", "num_workers"], got: {sorted(kwargs.keys())}'
            )
        # limit num_workers according to the number of CPUs
        if kwargs["num_workers"] > os.cpu_count():
            kwargs["num_workers"] = os.cpu_count()
            warnings.warn(f"`num_workers` limited to {os.cpu_count()}")
        # ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #
        # create dataloader
        return torch.utils.data.DataLoader(dataset=dataset, **{**default_kwargs, **kwargs})



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import warnings
from typing import Union

import lightning as L
from lightning.pytorch.utilities import CombinedLoader

from disent.dataset import DisentDataset
from disent.frameworks.ae import Ae
from disent.frameworks.vae import Vae

# ========================================================================= #
# Helper Functions                                                          #
# ========================================================================= #


def _get_dataset_and_ae_like(
    trainer_or_dataset: L.Trainer, pl_module: L.LightningModule, unwrap_groundtruth: bool = False
) -> (DisentDataset, Union[Ae, Vae]):
    assert isinstance(pl_module, (Ae, Vae)), f"{pl_module.__class__} is not an instance of {Ae} or {Vae}"
    # get dataset
    if isinstance(trainer_or_dataset, L.Trainer):
        trainer = trainer_or_dataset
        if hasattr(trainer, "datamodule") and (trainer.datamodule is not None):
            assert hasattr(
                trainer.datamodule, "dataset_train_noaug"
            )  # TODO: this is for experiments, another way of handling this should be added
            dataset = trainer.datamodule.dataset_train_noaug
        elif hasattr(trainer, "train_dataloader") and (trainer.train_dataloader is not None):
            if isinstance(trainer.train_dataloader, CombinedLoader):
                dataset = trainer.train_dataloader.loaders.dataset
            else:
                raise RuntimeError(f"invalid trainer.train_dataloader: {trainer.train_dataloader}")
        else:
            raise RuntimeError("could not retrieve dataset! please report this...")
    else:
        dataset = trainer_or_dataset
    # check dataset
    assert isinstance(dataset, DisentDataset), f"retrieved dataset is not an {DisentDataset.__name__}"
    # unwrap dataset
    if unwrap_groundtruth:
        if dataset.is_wrapped_gt_data:
            old_dataset, dataset = dataset, dataset.unwrapped_shallow_copy()
            warnings.warn(
                f"Unwrapped ground truth dataset returned! {type(old_dataset.data).__name__} -> {type(dataset.data).__name__}"
            )
    # done checks
    return dataset, pl_module


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


# class VaeLatentCorrelationLoggingCallback(BaseCallbackPeriodic):
#
#     def __init__(self, repeats_per_factor=10, every_n_steps=None, begin_first_step=False):
#         super().__init__(every_n_steps=every_n_steps, begin_first_step=begin_first_step)
#         self._repeats_per_factor = repeats_per_factor
#
#     def do_step(self, trainer: L.Trainer, pl_module: L.LightningModule):
#         # get dataset and vae framework from trainer and module
#         dataset, vae = _get_dataset_and_vae(trainer, pl_module)
#         # check if we need to skip
#         if not dataset.is_ground_truth:
#             warnings.warn(f'{dataset.__class__.__name__} is not an instance of {GroundTruthData.__name__}. Skipping callback: {self.__class__.__name__}!')
#             return
#         # TODO: CONVERT THIS TO A METRIC!
#         # log the correspondence between factors and the latent space.
#         num_samples = np.sum(dataset.ground_truth_data.factor_sizes) * self._repeats_per_factor
#         factors = dataset.ground_truth_data.sample_factors(num_samples)
#         # encode observations of factors
#         zs = np.concatenate([
#             to_numpy(vae.encode(dataset.dataset_batch_from_factors(factor_batch, mode='input').to(vae.device)))
#             for factor_batch in iter_chunks(factors, 256)
#         ])
#         z_size = zs.shape[-1]
#
#         # calculate correlation matrix
#         f_and_z = np.concatenate([factors.T, zs.T])
#         f_and_z_corr = np.corrcoef(f_and_z)
#         # get correlation submatricies
#         f_corr = f_and_z_corr[:z_size, :z_size]   # upper left
#         z_corr = f_and_z_corr[z_size:, z_size:]   # bottom right
#         fz_corr = f_and_z_corr[z_size:, :z_size]  # upper right | y: z, x: f
#         # get maximum z correlations per factor
#         z_to_f_corr_maxs = np.max(np.abs(fz_corr), axis=0)
#         f_to_z_corr_maxs = np.max(np.abs(fz_corr), axis=1)
#         assert len(z_to_f_corr_maxs) == z_size
#         assert len(f_to_z_corr_maxs) == dataset.ground_truth_data.num_factors
#         # average correlation
#         ave_f_to_z_corr = f_to_z_corr_maxs.mean()
#         ave_z_to_f_corr = z_to_f_corr_maxs.mean()
#
#         # print
#         log.info(f'ave latent correlation: {ave_z_to_f_corr}')
#         log.info(f'ave factor correlation: {ave_f_to_z_corr}')
#         # log everything
#         log_metrics(trainer.loggers, {
#             'metric.ave_latent_correlation': ave_z_to_f_corr,
#             'metric.ave_factor_correlation': ave_f_to_z_corr,
#         })
#         # make sure we only log the heatmap to WandB
#         wb_log_metrics(trainer.loggers, {
#             'metric.correlation_heatmap': wandb.plots.HeatMap(
#                 x_labels=[f'z{i}' for i in range(z_size)],
#                 y_labels=list(dataset.ground_truth_data.factor_names),
#                 matrix_values=fz_corr, show_text=False
#             ),
#         })
#
#         NUM = 1
#         # generate traversal value graphs
#         for i in range(z_size):
#             correlation = np.abs(f_corr[i, :])
#             correlation[i] = 0
#             for j in np.argsort(correlation)[::-1][:NUM]:
#                 if i == j:
#                     continue
#                 ix, iy = (i, j)  # if i < j else (j, i)
#                 plt.scatter(zs[:, ix], zs[:, iy])
#                 plt.title(f'z{ix}-vs-z{iy}')
#                 plt.xlabel(f'z{ix}')
#                 plt.ylabel(f'z{iy}')
#
#                 # wandb.log({f"chart.correlation.z{ix}-vs-z{iy}": plt})
#                 # make sure we only log to WANDB
#                 wb_log_metrics(trainer.loggers, {f"chart.correlation.z{ix}-vs-max-corr": plt})


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from typing import Literal
from typing import Optional
from typing import Tuple
from typing import Union

import lightning as L
import numpy as np
import torch

from disent.dataset import DisentDataset
from disent.frameworks.ae import Ae
from disent.frameworks.vae import Vae
from disent.util.lightning.callbacks._callbacks_base import BaseCallbackPeriodic
from disent.util.lightning.callbacks._helper import _get_dataset_and_ae_like
from disent.util.lightning.logger_util import wb_log_metrics
from disent.util.seeds import TempNumpySeed
from disent.util.visualize.vis_img import torch_to_images
from disent.util.visualize.vis_latents import make_decoded_latent_cycles
from disent.util.visualize.vis_util import make_animated_image_grid
from disent.util.visualize.vis_util import make_image_grid

log = logging.getLogger(__name__)


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


MinMaxHint = Optional[Union[int, Literal["auto"]]]
MeanStdHint = Optional[Union[Tuple[float, ...], float]]


def get_vis_min_max(
    recon_min: MinMaxHint = None,
    recon_max: MinMaxHint = None,
    recon_mean: MeanStdHint = None,
    recon_std: MeanStdHint = None,
) -> Union[Tuple[None, None], Tuple[np.ndarray, np.ndarray]]:
    # check recon_min and recon_max
    if (recon_min is not None) or (recon_max is not None):
        if (recon_mean is not None) or (recon_std is not None):
            raise ValueError("must choose either recon_min & recon_max OR recon_mean & recon_std, cannot specify both")
        if (recon_min is None) or (recon_max is None):
            raise ValueError("both recon_min & recon_max must be specified")
        # check strings
        if isinstance(recon_min, str) or isinstance(recon_max, str):
            if not (isinstance(recon_min, str) and isinstance(recon_max, str)):
                raise ValueError('both recon_min & recon_max must be "auto" if one is "auto"')
            return None, None  # "auto" -> None
    # check recon_mean and recon_std
    elif (recon_mean is not None) or (recon_std is not None):
        if (recon_min is not None) or (recon_max is not None):
            raise ValueError("must choose either recon_min & recon_max OR recon_mean & recon_std, cannot specify both")
        if (recon_mean is None) or (recon_std is None):
            raise ValueError("both recon_mean & recon_std must be specified")
        # set values:
        #  | ORIG: [0, 1]
        #  | TRANSFORM: (x - mean) / std         ->  [(0-mean)/std, (1-mean)/std]
        #  | REVERT:    (x - min) / (max - min)  ->  [0, 1]
        #  |            min=(0-mean)/std, max=(1-mean)/std
        recon_mean, recon_std = np.array(recon_mean, dtype="float32"), np.array(recon_std, dtype="float32")
        recon_min = np.divide(0 - recon_mean, recon_std)
        recon_max = np.divide(1 - recon_mean, recon_std)
    # set defaults
    if recon_min is None:
        recon_min = 0.0
    if recon_max is None:
        recon_max = 1.0
    # change type
    recon_min = np.array(recon_min)
    recon_max = np.array(recon_max)
    assert recon_min.ndim in (0, 1)
    assert recon_max.ndim in (0, 1)
    # checks
    assert np.all(recon_min < recon_max), f"recon_min={recon_min} must be less than recon_max={recon_max}"
    return recon_min, recon_max


# ========================================================================= #
# Latent Visualisation Callback                                             #
# ========================================================================= #


class VaeLatentCycleLoggingCallback(BaseCallbackPeriodic):
    def __init__(
        self,
        seed: Optional[int] = 7777,
        every_n_steps: Optional[int] = None,
        begin_first_step: bool = False,
        num_frames: int = 17,
        mode: str = "minmax_interval_cycle",
        num_stats_samples: int = 64,
        log_wandb: bool = True,  # TODO: detect this automatically?
        wandb_mode: str = "both",
        wandb_fps: int = 4,
        plt_show: bool = False,
        plt_block_size: float = 1.0,
        # recon_min & recon_max
        recon_min: MinMaxHint = None,  # scale data in this range [min, max] to [0, 1]
        recon_max: MinMaxHint = None,  # scale data in this range [min, max] to [0, 1]
        recon_mean: MeanStdHint = None,  # automatically converted to min & max [(0-mean)/std, (1-mean)/std], assuming original range of values is [0, 1]
        recon_std: MeanStdHint = None,  # automatically converted to min & max [(0-mean)/std, (1-mean)/std], assuming original range of values is [0, 1]
    ):
        super().__init__(every_n_steps, begin_first_step)
        self._seed = seed
        self._mode = mode
        self._num_stats_samples = num_stats_samples
        self._plt_show = plt_show
        self._plt_block_size = plt_block_size
        self._log_wandb = log_wandb
        self._wandb_mode = wandb_mode
        self._num_frames = num_frames
        self._fps = wandb_fps
        # checks
        assert wandb_mode in {
            "img",
            "vid",
            "both",
        }, f'invalid wandb_mode={repr(wandb_mode)}, must be one of: ("img", "vid", "both")'
        # normalize
        self._recon_min, self._recon_max = get_vis_min_max(
            recon_min=recon_min,
            recon_max=recon_max,
            recon_mean=recon_mean,
            recon_std=recon_std,
        )

    @torch.no_grad()
    def do_step(self, trainer: L.Trainer, pl_module: L.LightningModule):
        # exit early
        if not (self._plt_show or self._log_wandb):
            log.warning(f"skipping {self.__class__.__name__} neither `plt_show` or `log_wandb` is `True`!")
            return

        # feed forward and visualise everything!
        stills, animation, image = self.get_visualisations(trainer, pl_module)

        # log video -- none, img, vid, both
        # TODO: there might be a memory leak in making the video below? Or there could be one in the actual DSPRITES dataset? memory usage seems to be very high and increase on this dataset.
        if self._log_wandb:
            import wandb

            wandb_items = {}
            if self._wandb_mode in ("img", "both"):
                wandb_items[f"{self._mode}_img"] = wandb.Image(image)
            if self._wandb_mode in ("vid", "both"):
                wandb_items[f"{self._mode}_vid"] = (
                    wandb.Video(np.transpose(animation, [0, 3, 1, 2]), fps=self._fps, format="mp4"),
                )
            wb_log_metrics(trainer.loggers, wandb_items)

        # log locally
        if self._plt_show:
            from matplotlib import pyplot as plt

            fig, ax = plt.subplots(
                1, 1, figsize=(self._plt_block_size * stills.shape[1], self._plt_block_size * stills.shape[0])
            )
            ax.imshow(image)
            ax.axis("off")
            fig.tight_layout()
            plt.show()

    def get_visualisations(
        self,
        trainer_or_dataset: Union[L.Trainer, DisentDataset],
        pl_module: L.LightningModule,
    ) -> Union[
        Tuple[np.ndarray, np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray, torch.Tensor, torch.Tensor]
    ]:
        return self.generate_visualisations(
            trainer_or_dataset,
            pl_module,
            seed=self._seed,
            num_frames=self._num_frames,
            mode=self._mode,
            num_stats_samples=self._num_stats_samples,
            recon_min=self._recon_min,
            recon_max=self._recon_max,
            recon_mean=None,
            recon_std=None,
        )

    @classmethod
    def generate_visualisations(
        cls,
        trainer_or_dataset: Union[L.Trainer, DisentDataset],
        pl_module: L.LightningModule,
        seed: Optional[int] = 7777,
        num_frames: int = 17,
        mode: str = "fitted_gaussian_cycle",
        num_stats_samples: int = 64,
        # recon_min & recon_max
        recon_min: MinMaxHint = None,
        recon_max: MinMaxHint = None,
        recon_mean: MeanStdHint = None,
        recon_std: MeanStdHint = None,
    ) -> Union[
        Tuple[np.ndarray, np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray, torch.Tensor, torch.Tensor]
    ]:
        # normalize
        recon_min, recon_max = get_vis_min_max(
            recon_min=recon_min,
            recon_max=recon_max,
            recon_mean=recon_mean,
            recon_std=recon_std,
        )

        # get dataset and vae framework from trainer and module
        dataset, vae = _get_dataset_and_ae_like(trainer_or_dataset, pl_module, unwrap_groundtruth=True)

        # ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #
        # generate traversal
        # ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #

        # get random sample of z_means and z_logvars for computing the range of values for the latent_cycle
        with TempNumpySeed(seed):
            batch, indices = dataset.dataset_sample_batch(
                num_stats_samples, mode="input", replace=True, return_indices=True
            )  # replace just in case the dataset it tiny
            batch = batch.to(vae.device)

        # get representations
        if isinstance(vae, Vae):
            # variational auto-encoder
            ds_posterior, ds_prior = vae.encode_dists(batch)
            zs_mean, zs_logvar = ds_posterior.mean, torch.log(ds_posterior.variance)
        elif isinstance(vae, Ae):
            # auto-encoder
            zs_mean = vae.encode(batch)
            zs_logvar = torch.ones_like(zs_mean)
        else:
            log.warning(
                f"cannot run {cls.__name__}, unsupported type: {type(vae)}, must be {Ae.__name__} or {Vae.__name__}"
            )
            return

        # get min and max if auto
        if (recon_min is None) or (recon_max is None):
            if recon_min is None:
                recon_min = float(torch.amin(batch).cpu())
            if recon_max is None:
                recon_max = float(torch.amax(batch).cpu())
            log.info(f"auto visualisation min: {recon_min} and max: {recon_max} obtained from {len(batch)} samples")

        # produce latent cycle still images & convert them to images
        stills = make_decoded_latent_cycles(
            vae.decode,
            zs_mean,
            zs_logvar,
            mode=mode,
            num_animations=1,
            num_frames=num_frames,
            decoder_device=vae.device,
        )[0]
        stills = torch_to_images(
            stills, in_dims="CHW", out_dims="HWC", in_min=recon_min, in_max=recon_max, always_rgb=True, to_numpy=True
        )

        # generate the video frames and image grid from the stills
        # - TODO: this needs to be fixed to not use logvar, but rather the representations or distributions themselves
        # - TODO: should this not use `visualize_dataset_traversal`?
        frames = make_animated_image_grid(stills, pad=4, border=True, bg_color=None)
        image = make_image_grid(
            stills.reshape(-1, *stills.shape[2:]), num_cols=stills.shape[1], pad=4, border=True, bg_color=None
        )

        # done
        return stills, frames, image


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from torch import Tensor
from torch import nn

from disent.model import DisentDecoder
from disent.model import DisentEncoder

# ========================================================================= #
# disentanglement_lib Conv models                                           #
# ========================================================================= #


class EncoderConv64(DisentEncoder):
    """
    Convolutional encoder used in beta-VAE paper for the chairs data.
    Based on row 4-6 of Table 1 on page 13 of "beta-VAE: Learning Basic Visual
    Concepts with a Constrained Variational Framework"
    (https://openreview.net/forum?id=Sy2fzU9gl)

    Reference Implementation:
        - https://github.com/google-research/disentanglement_lib/blob/master/disentanglement_lib/methods/shared/architectures.py
        >>> def conv_encoder(input_tensor, num_latent):
        >>>     e1 = tf.layers.conv2d(inputs=input_tensor, filters=32, kernel_size=4, strides=2, activation=tf.nn.relu, padding="same", name="e1",)
        >>>     e2 = tf.layers.conv2d(inputs=e1,           filters=32, kernel_size=4, strides=2, activation=tf.nn.relu, padding="same", name="e2",)
        >>>     e3 = tf.layers.conv2d(inputs=e2,           filters=64, kernel_size=2, strides=2, activation=tf.nn.relu, padding="same", name="e3",)  # TODO: this does not match beta-vae paper
        >>>     e4 = tf.layers.conv2d(inputs=e3,           filters=64, kernel_size=2, strides=2, activation=tf.nn.relu, padding="same", name="e4",)  # TODO: this does not match beta-vae paper
        >>>     flat_e4 = tf.layers.flatten(e4)
        >>>     e5      = tf.layers.dense(flat_e4, 256,        activation=tf.nn.relu, name="e5")
        >>>     means   = tf.layers.dense(e5,      num_latent, activation=None,       name="means")
        >>>     log_var = tf.layers.dense(e5,      num_latent, activation=None,       name="log_var")
        >>>     return means, log_var
    """

    def __init__(self, x_shape=(3, 64, 64), z_size=6, z_multiplier=1):
        # checks
        (C, H, W) = x_shape
        assert (H, W) == (64, 64), "This model only works with image size 64x64."
        super().__init__(x_shape=x_shape, z_size=z_size, z_multiplier=z_multiplier)

        self.model = nn.Sequential(
            nn.Conv2d(in_channels=C, out_channels=32, kernel_size=4, stride=2, padding=2),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels=32, out_channels=32, kernel_size=4, stride=2, padding=2),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_channels=32, out_channels=64, kernel_size=4, stride=2, padding=2),
            nn.ReLU(inplace=True),  # This was reverted to kernel size 4x4 from 2x2, to match beta-vae paper
            nn.Conv2d(in_channels=64, out_channels=64, kernel_size=4, stride=2, padding=2),
            nn.ReLU(inplace=True),  # This was reverted to kernel size 4x4 from 2x2, to match beta-vae paper
            nn.Flatten(),
            nn.Linear(in_features=1600, out_features=256),
            nn.ReLU(inplace=True),
            nn.Linear(
                in_features=256, out_features=self.z_total
            ),  # we combine the two networks in the reference implementation and use torch.chunk(2, dim=-1) to get mu & logvar
        )

    def encode(self, x) -> (Tensor, Tensor):
        return self.model(x)


class DecoderConv64(DisentDecoder):
    """
    Convolutional decoder used in beta-VAE paper for the chairs data.
    Based on row 3 of Table 1 on page 13 of "beta-VAE: Learning Basic Visual
    Concepts with a Constrained Variational Framework"
    (https://openreview.net/forum?id=Sy2fzU9gl)

    Reference Implementation:
        - https://github.com/google-research/disentanglement_lib/blob/master/disentanglement_lib/methods/shared/architectures.py
        >>> def deconv_decoder(latent_tensor, output_shape):
        >>>     d1 = tf.layers.dense(latent_tensor,  256, activation=tf.nn.relu)
        >>>     d2 = tf.layers.dense(d1,            1024, activation=tf.nn.relu)
        >>>     d2_reshaped = tf.reshape(d2, shape=[-1, 4, 4, 64])
        >>>     d3 = tf.layers.conv2d_transpose(inputs=d2_reshaped, filters=64,              kernel_size=4, strides=2, activation=tf.nn.relu, padding="same")
        >>>     d4 = tf.layers.conv2d_transpose(inputs=d3,          filters=32,              kernel_size=4, strides=2, activation=tf.nn.relu, padding="same")
        >>>     d5 = tf.layers.conv2d_transpose(inputs=d4,          filters=32,              kernel_size=4, strides=2, activation=tf.nn.relu, padding="same")
        >>>     d6 = tf.layers.conv2d_transpose(inputs=d5,          filters=output_shape[2], kernel_size=4, strides=2,                        padding="same")
        >>>     return tf.reshape(d6, [-1] + output_shape)
    """

    def __init__(self, x_shape=(3, 64, 64), z_size=6, z_multiplier=1):
        (C, H, W) = x_shape
        assert (H, W) == (64, 64), "This model only works with image size 64x64."
        super().__init__(x_shape=x_shape, z_size=z_size, z_multiplier=z_multiplier)

        self.model = nn.Sequential(
            nn.Linear(in_features=self.z_size, out_features=256),
            nn.ReLU(inplace=True),
            nn.Linear(in_features=256, out_features=1024),
            nn.ReLU(inplace=True),
            nn.Unflatten(dim=1, unflattened_size=[64, 4, 4]),
            nn.ConvTranspose2d(in_channels=64, out_channels=64, kernel_size=4, stride=2, padding=1),
            nn.ReLU(inplace=True),
            nn.ConvTranspose2d(in_channels=64, out_channels=32, kernel_size=4, stride=2, padding=1),
            nn.ReLU(inplace=True),
            nn.ConvTranspose2d(in_channels=32, out_channels=32, kernel_size=4, stride=2, padding=1),
            nn.ReLU(inplace=True),
            nn.ConvTranspose2d(in_channels=32, out_channels=C, kernel_size=4, stride=2, padding=1),
        )

    def decode(self, z) -> Tensor:
        return self.model(z)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os

import numpy as np
import torch


"""
helpful functions that do not fit nicely into any other file.
"""

log = logging.getLogger(__name__)

# ========================================================================= #
# seeds                                                                     #
# ========================================================================= #


def is_test_run():
    """
    This is used internally to test some scripts. There is no need
    to use this function in your own scripts.
    """
    return bool(os.environ.get("DISENT_TEST_RUN", False))


def _set_test_run():
    os.environ["DISENT_TEST_RUN"] = "True"


# ========================================================================= #
# Conversion                                                                #
# ========================================================================= #


def to_numpy(array) -> np.ndarray:
    """
    Handles converting any array like object to a numpy array.
    specifically with support for a tensor
    """
    # TODO: replace... maybe with kornia
    if torch.is_tensor(array):
        return array.cpu().detach().numpy()
    # recursive conversion
    # not super efficient but allows handling of PIL.Image and other nested data.
    elif isinstance(array, (list, tuple)):
        return np.stack([to_numpy(elem) for elem in array], axis=0)
    else:
        return np.array(array)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Optional

import numpy as np

from disent.dataset.data import GroundTruthData
from disent.dataset.sampling._base import BaseDisentSampler
from disent.dataset.util.state_space import StateSpace


class GroundTruthPairOrigSampler(BaseDisentSampler):
    def uninit_copy(self) -> "GroundTruthPairOrigSampler":
        return GroundTruthPairOrigSampler(p_k=self.p_k)

    def __init__(
        self,
        # num_differing_factors
        p_k: int = 1,
    ):
        """
        Sampler that emulates choosing factors like:
        https://github.com/google-research/disentanglement_lib/blob/master/disentanglement_lib/methods/weak/train_weak_lib.py
        """
        super().__init__(num_samples=2)
        # DIFFERING FACTORS
        self.p_k = p_k
        # dataset variable
        self._state_space: Optional[StateSpace] = None

    def _init(self, dataset):
        assert isinstance(
            dataset, GroundTruthData
        ), f"dataset must be an instance of {repr(GroundTruthData.__class__.__name__)}, got: {repr(dataset)}"
        self._state_space = dataset.state_space_copy()

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # CORE                                                                  #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    def _sample_idx(self, idx):
        f0, f1 = self.datapoint_sample_factors_pair(idx)
        return (
            self._state_space.pos_to_idx(f0),
            self._state_space.pos_to_idx(f1),
        )

    def datapoint_sample_factors_pair(self, idx):
        """
        This function is based on _sample_weak_pair_factors()
        Except deterministic for the first item in the pair, based off of idx.
        """
        # randomly sample the first observation -- In our case we just use the idx
        sampled_factors = self._state_space.idx_to_pos(idx)
        # sample the next observation with k differing factors
        next_factors, k = _sample_k_differing(sampled_factors, self._state_space, k=self.p_k)
        # return the samples
        return sampled_factors, next_factors


def _sample_k_differing(factors, state_space: StateSpace, k=1):
    """
    Resample the factors used for the corresponding item in a pair.
      - Based on simple_dynamics() from:
        https://github.com/google-research/disentanglement_lib/blob/master/disentanglement_lib/methods/weak/train_weak_lib.py
    """
    # checks for factors
    factors = np.array(factors)
    assert factors.ndim == 1
    # sample k
    if k <= 0:
        k = np.random.randint(1, state_space.num_factors)
    # randomly choose 1 or k
    # TODO: This is in disentanglement lib, HOWEVER is this not a mistake?
    #       A bug report has been submitted to disentanglement_lib for clarity:
    #       https://github.com/google-research/disentanglement_lib/issues/31
    k = np.random.choice([1, k])
    # generate list of differing indices
    index_list = np.random.choice(len(factors), k, replace=False)
    # randomly update factors
    for index in index_list:
        factors[index] = np.random.choice(state_space.factor_sizes[index])
    # return!
    return factors, k


def _sample_weak_pair_factors(state_space: StateSpace):  # pragma: no cover
    """
    Sample a weakly supervised pair from the given GroundTruthData.
      - Based on weak_dataset_generator() from:
        https://github.com/google-research/disentanglement_lib/blob/master/disentanglement_lib/methods/weak/train_weak_lib.py
    """
    # randomly sample the first observation
    sampled_factors = state_space.sample_factors(1)
    # sample the next observation with k differing factors
    next_factors, k = _sample_k_differing(sampled_factors, state_space, k=1)
    # return the samples
    return sampled_factors, next_factors



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from typing import Callable
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import lightning as L
import numpy as np
import torch

import disent.util.strings.colors as c
from disent.dataset import DisentDataset
from disent.dataset.data import GroundTruthData
from disent.frameworks.ae import Ae
from disent.frameworks.vae import Vae
from disent.util.function import wrapped_partial
from disent.util.iters import chunked
from disent.util.lightning.callbacks._callbacks_base import BaseCallbackPeriodic
from disent.util.lightning.callbacks._helper import _get_dataset_and_ae_like
from disent.util.lightning.logger_util import wb_log_metrics
from disent.util.profiling import Timer
from disent.util.seeds import TempNumpySeed
from disent.util.visualize.plot import plt_subplots_imshow

log = logging.getLogger(__name__)


# ========================================================================= #
# Helper Functions                                                          #
# ========================================================================= #


# helper
def _to_dmat(
    size: int,
    i_a: np.ndarray,
    i_b: np.ndarray,
    dists: Union[torch.Tensor, np.ndarray],
) -> np.ndarray:
    if isinstance(dists, torch.Tensor):
        dists = dists.detach().cpu().numpy()
    # checks
    assert i_a.ndim == 1
    assert i_a.shape == i_b.shape
    assert i_a.shape == dists.shape
    # compute
    dmat = np.zeros([size, size], dtype="float32")
    dmat[i_a, i_b] = dists
    dmat[i_b, i_a] = dists
    return dmat


_AE_DIST_NAMES = ("x", "z", "x_recon")
_VAE_DIST_NAMES = ("x", "z", "kl", "x_recon")


@torch.no_grad()
def _get_dists_ae(ae: Ae, x_a: torch.Tensor, x_b: torch.Tensor):
    # feed forware
    z_a, z_b = ae.encode(x_a), ae.encode(x_b)
    r_a, r_b = ae.decode(z_a), ae.decode(z_b)
    # distances
    return [
        ae.recon_handler.compute_pairwise_loss(x_a, x_b),
        torch.norm(z_a - z_b, p=2, dim=-1),  # l2 dist
        ae.recon_handler.compute_pairwise_loss(r_a, r_b),
    ]


@torch.no_grad()
def _get_dists_vae(vae: Vae, x_a: torch.Tensor, x_b: torch.Tensor):
    from torch.distributions import kl_divergence

    # feed forward
    (z_post_a, z_prior_a), (z_post_b, z_prior_b) = vae.encode_dists(x_a), vae.encode_dists(x_b)
    z_a, z_b = z_post_a.mean, z_post_b.mean
    r_a, r_b = vae.decode(z_a), vae.decode(z_b)
    # dists
    kl_ab = 0.5 * kl_divergence(z_post_a, z_post_b) + 0.5 * kl_divergence(z_post_b, z_post_a)
    # distances
    return [
        vae.recon_handler.compute_pairwise_loss(x_a, x_b),
        torch.norm(z_a - z_b, p=2, dim=-1),  # l2 dist
        vae.recon_handler._pairwise_reduce(kl_ab),
        vae.recon_handler.compute_pairwise_loss(r_a, r_b),
    ]


def _get_dists_fn(
    model: Ae,
) -> Tuple[Optional[Tuple[str, ...]], Optional[Callable[[object, object], Sequence[Sequence[float]]]]]:
    # get aggregate function
    if isinstance(model, Vae):
        dists_names, dists_fn = _VAE_DIST_NAMES, wrapped_partial(_get_dists_vae, model)
    elif isinstance(model, Ae):
        dists_names, dists_fn = _AE_DIST_NAMES, wrapped_partial(_get_dists_ae, model)
    else:
        dists_names, dists_fn = None, None
    return dists_names, dists_fn


@torch.no_grad()
def _collect_dists_subbatches(
    dists_fn: Callable[[object, object], Sequence[Sequence[float]]],
    batch: torch.Tensor,
    i_a: np.ndarray,
    i_b: np.ndarray,
    batch_size: int = 64,
):
    # feed forward
    results = []
    for idxs in chunked(np.stack([i_a, i_b], axis=-1), chunk_size=batch_size):
        ia, ib = idxs.T
        x_a, x_b = batch[ia], batch[ib]
        # feed forward
        data = dists_fn(x_a, x_b)
        results.append(data)
    return [torch.cat(r, dim=0) for r in zip(*results)]


def _compute_and_collect_dists(
    dataset: DisentDataset,
    dists_fn,
    dists_names: Sequence[str],
    traversal_repeats: int = 100,
    batch_size: int = 32,
    include_gt_factor_dists: bool = True,
    transform_batch: Callable[[object], object] = None,
    data_mode: str = "input",
) -> Tuple[Tuple[str, ...], List[List[np.ndarray]]]:
    assert traversal_repeats > 0
    gt_data = dataset.gt_data
    # generate
    f_grid = []
    # generate
    for f_idx, f_size in enumerate(gt_data.factor_sizes):
        # save for the current factor (traversal_repeats, len(names), len(i_a))
        f_dists = []
        # upper triangle excluding diagonal
        i_a, i_b = np.triu_indices(f_size, k=1)
        # repeat over random traversals
        for i in range(traversal_repeats):
            # get random factor traversal
            factors = gt_data.sample_random_factor_traversal(f_idx=f_idx)
            indices = gt_data.pos_to_idx(factors)
            # load data
            batch = dataset.dataset_batch_from_indices(indices, data_mode)
            if transform_batch is not None:
                batch = transform_batch(batch)
            # feed forward & compute dists -- (len(names), len(i_a))
            dists = _collect_dists_subbatches(dists_fn=dists_fn, batch=batch, i_a=i_a, i_b=i_b, batch_size=batch_size)
            assert len(dists) == len(dists_names)
            # distances
            f_dists.append(dists)
        # aggregate all dists into distances matrices for current factor
        f_dmats = [
            _to_dmat(size=f_size, i_a=i_a, i_b=i_b, dists=torch.stack(dists, dim=0).mean(dim=0))
            for dists in zip(*f_dists)
        ]
        # handle factors
        if include_gt_factor_dists:
            i_dmat = _to_dmat(size=f_size, i_a=i_a, i_b=i_b, dists=np.abs(factors[i_a] - factors[i_b]).sum(axis=-1))
            f_dmats = [i_dmat, *f_dmats]
        # append data
        f_grid.append(f_dmats)
    # handle factors
    if include_gt_factor_dists:
        dists_names = ("factors", *dists_names)
    # done
    return tuple(dists_names), f_grid


def compute_factor_distances(
    dataset: DisentDataset,
    dists_fn,
    dists_names: Sequence[str],
    traversal_repeats: int = 100,
    batch_size: int = 32,
    include_gt_factor_dists: bool = True,
    transform_batch: Callable[[object], object] = None,
    seed: Optional[int] = 777,
    data_mode: str = "input",
) -> Tuple[Tuple[str, ...], List[List[np.ndarray]]]:
    # log this callback
    gt_data = dataset.gt_data
    log.info(f"| {gt_data.name} - computing factor distances...")
    # compute various distances matrices for each factor
    with Timer() as timer, TempNumpySeed(seed):
        dists_names, f_grid = _compute_and_collect_dists(
            dataset=dataset,
            dists_fn=dists_fn,
            dists_names=dists_names,
            traversal_repeats=traversal_repeats,
            batch_size=batch_size,
            include_gt_factor_dists=include_gt_factor_dists,
            transform_batch=transform_batch,
            data_mode=data_mode,
        )
    # log this callback!
    log.info(f"| {gt_data.name} - computed factor distances! time{c.GRY}={c.lYLW}{timer.pretty:<9}{c.RST}")
    return dists_names, f_grid


def plt_factor_distances(
    gt_data: GroundTruthData,
    f_grid: List[List[np.ndarray]],
    dists_names: Sequence[str],
    title: str,
    plt_block_size: float = 1.25,
    plt_transpose: bool = False,
    plt_cmap="Blues",
):
    # plot information
    imshow_kwargs = dict(cmap=plt_cmap)
    figsize = (plt_block_size * len(f_grid[0]), plt_block_size * gt_data.num_factors)
    # plot!
    if not plt_transpose:
        fig, axs = plt_subplots_imshow(
            grid=f_grid,
            col_labels=dists_names,
            row_labels=gt_data.factor_names,
            figsize=figsize,
            title=title,
            imshow_kwargs=imshow_kwargs,
        )
    else:
        fig, axs = plt_subplots_imshow(
            grid=list(zip(*f_grid)),
            col_labels=gt_data.factor_names,
            row_labels=dists_names,
            figsize=figsize[::-1],
            title=title,
            imshow_kwargs=imshow_kwargs,
        )
    # done
    return fig, axs


# ========================================================================= #
# Data Dists Visualisation Callback                                         #
# ========================================================================= #


class VaeGtDistsLoggingCallback(BaseCallbackPeriodic):
    def __init__(
        self,
        seed: Optional[int] = 7777,
        every_n_steps: Optional[int] = None,
        traversal_repeats: int = 100,
        begin_first_step: bool = False,
        plt_block_size: float = 1.25,
        plt_show: bool = False,
        plt_transpose: bool = False,
        log_wandb: bool = True,  # TODO: detect this automatically?
        batch_size: int = 128,
        include_factor_dists: bool = True,
    ):
        assert traversal_repeats > 0
        self._traversal_repeats = traversal_repeats
        self._seed = seed
        self._plt_block_size = plt_block_size
        self._plt_show = plt_show
        self._log_wandb = log_wandb
        self._include_gt_factor_dists = include_factor_dists
        self._transpose_plot = plt_transpose
        self._batch_size = batch_size
        super().__init__(every_n_steps, begin_first_step)

    @torch.no_grad()
    def do_step(self, trainer: L.Trainer, pl_module: L.LightningModule):
        # exit early
        if not (self._plt_show or self._log_wandb):
            log.warning(f"skipping {self.__class__.__name__} neither `plt_show` or `log_wandb` is `True`!")
            return
        # get dataset and vae framework from trainer and module
        dataset, vae = _get_dataset_and_ae_like(trainer, pl_module, unwrap_groundtruth=True)
        # exit early
        if not dataset.is_ground_truth:
            log.warning(f"cannot run {self.__class__.__name__} over non-ground-truth data, skipping!")
            return
        # get aggregate function
        dists_names, dists_fn = _get_dists_fn(vae)
        if (dists_names is None) or (dists_fn is None):
            log.warning(
                f"cannot run {self.__class__.__name__}, unsupported model type: {type(vae)}, must be {Ae.__name__} or {Vae.__name__}"
            )
            return
        # compute various distances matrices for each factor
        dists_names, f_grid = compute_factor_distances(
            dataset=dataset,
            dists_fn=dists_fn,
            dists_names=dists_names,
            traversal_repeats=self._traversal_repeats,
            batch_size=self._batch_size,
            include_gt_factor_dists=self._include_gt_factor_dists,
            transform_batch=lambda batch: batch.to(vae.device),
            seed=self._seed,
            data_mode="input",
        )
        # plot these results
        fig, axs = plt_factor_distances(
            gt_data=dataset.gt_data,
            f_grid=f_grid,
            dists_names=dists_names,
            title=f"{vae.__class__.__name__}: {dataset.gt_data.name.capitalize()} Distances",
            plt_block_size=self._plt_block_size,
            plt_transpose=self._transpose_plot,
            plt_cmap="Blues",
        )
        # show the plot
        if self._plt_show:
            import matplotlib.pyplot as plt

            plt.show()
        # log the plot to wandb
        if self._log_wandb:
            import wandb

            wb_log_metrics(trainer.loggers, {"factor_distances": wandb.Image(fig)})


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~


import warnings
from functools import lru_cache
from numbers import Number
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import torch
from numpy.core.multiarray import normalize_axis_index
from PIL import Image

# ========================================================================= #
# Type Hints                                                                #
# ========================================================================= #


# from torch.testing._internal.common_utils import numpy_to_torch_dtype_dict
_NP_TO_TORCH_DTYPE = {
    np.dtype("bool"): torch.bool,
    np.dtype("uint8"): torch.uint8,
    np.dtype("int8"): torch.int8,
    np.dtype("int16"): torch.int16,
    np.dtype("int32"): torch.int32,
    np.dtype("int64"): torch.int64,
    np.dtype("float16"): torch.float16,
    np.dtype("float32"): torch.float32,
    np.dtype("float64"): torch.float64,
    np.dtype("complex64"): torch.complex64,
    np.dtype("complex128"): torch.complex128,
}


MinMaxHint = Union[Number, Tuple[Number, ...], np.ndarray]


@lru_cache()
def _dtype_min_max(dtype: torch.dtype) -> Tuple[Union[float, int], Union[float, int]]:
    """Get the min and max values for a dtype"""
    dinfo = torch.finfo(dtype) if dtype.is_floating_point else torch.iinfo(dtype)
    return dinfo.min, dinfo.max


@lru_cache()
def _check_image_dtype(dtype: torch.dtype):
    """Check that a dtype can hold image values"""
    # check that the datatype is within the right range -- this is not actually necessary if the below is correct!
    dmin, dmax = _dtype_min_max(dtype)
    imin, imax = (0, 1) if dtype.is_floating_point else (0, 255)
    assert (dmin <= imin) and (
        imax <= dmax
    ), f"The dtype: {repr(dtype)} with range [{dmin}, {dmax}] cannot store image values in the range [{imin}, {imax}]"
    # check the datatype is allowed
    if dtype not in _ALLOWED_DTYPES:
        raise TypeError(f"The dtype: {repr(dtype)} is not allowed, must be one of: {list(_ALLOWED_DTYPES)}")
    # return the min and max values
    return imin, imax


# ========================================================================= #
# Image Helper Functions                                                    #
# ========================================================================= #


def torch_image_has_valid_range(tensor: torch.Tensor, check_mode: Optional[str] = None) -> bool:
    """
    Check that the range of values in the image is correct!
    """
    if check_mode not in {"error", "warn", "bool", None}:
        raise KeyError(f"invalid check_mode: {repr(check_mode)}")
    # get the range for the dtype
    imin, imax = _check_image_dtype(tensor.dtype)
    # get the values
    m = tensor.amin().cpu().numpy()
    M = tensor.amax().cpu().numpy()
    if (m < imin) or (imax < M):
        if check_mode == "error":
            raise ValueError(
                f"images value range: [{m}, {M}] is outside of the required range: [{imin}, {imax}], for dtype: {repr(tensor.dtype)}"
            )
        elif check_mode == "warn":
            warnings.warn(
                f"images value range: [{m}, {M}] is outside of the required range: [{imin}, {imax}], for dtype: {repr(tensor.dtype)}"
            )
        return False
    return True


@torch.no_grad()
def torch_image_clamp(tensor: torch.Tensor, clamp_mode: str = "warn") -> torch.Tensor:
    """
    Clamp the image based on the dtype
    Valid `clamp_mode`s are {'warn', 'error', 'clamp'}
    """
    # check range of values
    if clamp_mode in ("warn", "error"):
        torch_image_has_valid_range(tensor, check_mode=clamp_mode)
    elif clamp_mode != "clamp":
        raise KeyError(f"invalid clamp mode: {repr(clamp_mode)}")
    # get the range for the dtype
    imin, imax = _check_image_dtype(tensor.dtype)
    # clamp!
    return torch.clamp(tensor, imin, imax)


@torch.no_grad()
def torch_image_to_dtype(tensor: torch.Tensor, out_dtype: torch.dtype):
    """
    Convert an image to the specified dtype
    - Scaling is automatically performed based on the input and output dtype
      Floats should be in the range [0, 1], integers should be in the range [0, 255]
    - if precision will be lost (), then the values are clamped!
    """
    _check_image_dtype(tensor.dtype)
    _check_image_dtype(out_dtype)
    # check scale
    torch_image_has_valid_range(tensor, check_mode="error")
    # convert
    if tensor.dtype.is_floating_point and (not out_dtype.is_floating_point):
        # [float -> int] -- cast after scaling
        return torch.clamp(tensor * 255, 0, 255).to(out_dtype)
    elif (not tensor.dtype.is_floating_point) and out_dtype.is_floating_point:
        # [int -> float] -- cast before scaling
        return torch.clamp(tensor.to(out_dtype) / 255, 0, 1)
    else:
        # [int -> int] | [float -> float]
        return tensor.to(out_dtype)


@torch.no_grad()
def torch_image_normalize_channels(
    tensor: torch.Tensor,
    in_min: MinMaxHint,
    in_max: MinMaxHint,
    channel_dim: int = -1,
    out_dtype: Optional[torch.dtype] = None,
):
    if out_dtype is None:
        out_dtype = tensor.dtype
    # check dtypes
    _check_image_dtype(out_dtype)
    assert out_dtype.is_floating_point, f"out_dtype must be a floating point, got: {repr(out_dtype)}"
    # get norm values padded to the dimension of the channel
    in_min, in_max = _torch_channel_broadcast_scale_values(
        in_min, in_max, in_dtype=tensor.dtype, dim=channel_dim, ndim=tensor.ndim
    )
    # convert
    tensor = tensor.to(out_dtype)
    in_min = torch.as_tensor(in_min, dtype=tensor.dtype, device=tensor.device)
    in_max = torch.as_tensor(in_max, dtype=tensor.dtype, device=tensor.device)
    # warn if the values are the same
    if torch.any(in_min == in_max):
        m = in_min.cpu().detach().numpy()
        M = in_min.cpu().detach().numpy()
        warnings.warn(f"minimum: {m} and maximum: {M} values are the same, scaling values to zero.")
    # handle equal values
    divisor = in_max - in_min
    divisor[divisor == 0] = 1
    # normalize
    return (tensor - in_min) / divisor


# ========================================================================= #
# Argument Helper                                                           #
# ========================================================================= #


# float16 doesnt always work, rather convert to float32 first
_ALLOWED_DTYPES = {
    torch.float32,
    torch.float64,
    torch.uint8,
    torch.int,
    torch.int16,
    torch.int32,
    torch.int64,
    torch.long,
}


@lru_cache()
def _torch_to_images_normalise_args(
    in_tensor_shape: Tuple[int, ...],
    in_tensor_dtype: torch.dtype,
    in_dims: str,
    out_dims: str,
    in_dtype: Optional[torch.dtype],
    out_dtype: Optional[torch.dtype],
):
    # check types
    if not isinstance(in_dims, str):
        raise TypeError(f"in_dims must be of type: {str}, but got: {type(in_dims)}")
    if not isinstance(out_dims, str):
        raise TypeError(f"out_dims must be of type: {str}, but got: {type(out_dims)}")
    # normalise dim names
    in_dims = in_dims.upper()
    out_dims = out_dims.upper()
    # check dim values
    if sorted(in_dims) != sorted("CHW"):
        raise KeyError(
            f'in_dims contains the symbols: {repr(in_dims)}, must contain only permutations of: {repr("CHW")}'
        )
    if sorted(out_dims) != sorted("CHW"):
        raise KeyError(
            f'out_dims contains the symbols: {repr(out_dims)}, must contain only permutations of: {repr("CHW")}'
        )
    # get dimension indices
    in_c_dim = in_dims.index("C") - len(in_dims)
    out_c_dim = out_dims.index("C") - len(out_dims)
    transpose_indices = tuple(in_dims.index(c) - len(in_dims) for c in out_dims)
    # check image tensor
    if len(in_tensor_shape) < 3:
        raise ValueError(
            f'images must have 3 or more dimensions corresponding to: (..., {", ".join(in_dims)}), but got shape: {in_tensor_shape}'
        )
    if in_tensor_shape[in_c_dim] not in (1, 3):
        raise ValueError(
            f'images do not have the correct number of channels for dim "C", required: 1 or 3. Input format is (..., {", ".join(in_dims)}), but got shape: {in_tensor_shape}'
        )
    # get default values
    if in_dtype is None:
        in_dtype = in_tensor_dtype
    if out_dtype is None:
        out_dtype = in_dtype
    # check dtypes allowed
    if in_dtype not in _ALLOWED_DTYPES:
        raise TypeError(f"in_dtype is not allowed, got: {repr(in_dtype)} must be one of: {list(_ALLOWED_DTYPES)}")
    if out_dtype not in _ALLOWED_DTYPES:
        raise TypeError(f"out_dtype is not allowed, got: {repr(out_dtype)} must be one of: {list(_ALLOWED_DTYPES)}")
    # done!
    return transpose_indices, in_dtype, out_dtype, out_c_dim


def _torch_channel_broadcast_scale_values(
    in_min: MinMaxHint,
    in_max: MinMaxHint,
    in_dtype: torch.dtype,
    dim: int,
    ndim: int,
) -> Tuple[List[Number], List[Number]]:
    return __torch_channel_broadcast_scale_values(
        in_min=tuple(np.array(in_min).reshape(-1).tolist()),  # TODO: this is slow?
        in_max=tuple(np.array(in_max).reshape(-1).tolist()),  # TODO: this is slow?
        in_dtype=in_dtype,
        dim=dim,
        ndim=ndim,
    )


@lru_cache()
@torch.no_grad()
def __torch_channel_broadcast_scale_values(
    in_min: MinMaxHint,
    in_max: MinMaxHint,
    in_dtype: torch.dtype,
    dim: int,
    ndim: int,
) -> Tuple[List[Number], List[Number]]:
    # get the default values
    in_min: np.ndarray = np.array((0.0 if in_dtype.is_floating_point else 0.0) if (in_min is None) else in_min)
    in_max: np.ndarray = np.array((1.0 if in_dtype.is_floating_point else 255.0) if (in_max is None) else in_max)
    # add missing axes
    if in_min.ndim == 0:
        in_min = in_min[None]
    if in_max.ndim == 0:
        in_max = in_max[None]
    # checks
    assert in_min.ndim == 1
    assert in_max.ndim == 1
    assert np.all(in_min <= in_max), f"min values are not <= the max values: {in_min} !<= {in_max}"
    # normalize dim
    dim = normalize_axis_index(dim, ndim=ndim)
    # pad dim
    r_pad = ndim - (dim + 1)
    if r_pad > 0:
        in_min = in_min[(...,) + ((None,) * r_pad)]
        in_max = in_max[(...,) + ((None,) * r_pad)]
    # done!
    return in_min.tolist(), in_max.tolist()


# ========================================================================= #
# Image Conversion                                                          #
# ========================================================================= #


@torch.no_grad()
def torch_to_images(
    tensor: torch.Tensor,
    in_dims: str = "CHW",  # we always treat numpy by default as HWC, and torch.Tensor as CHW
    out_dims: str = "HWC",
    in_dtype: Optional[torch.dtype] = None,
    out_dtype: Optional[torch.dtype] = torch.uint8,
    clamp_mode: str = "warn",  # clamp, warn, error
    always_rgb: bool = False,
    in_min: Optional[MinMaxHint] = None,
    in_max: Optional[MinMaxHint] = None,
    to_numpy: bool = False,
) -> Union[torch.Tensor, np.ndarray]:
    """
    Convert a batch of image-like tensors to images.
    A batch in this case consists of an arbitrary number of dimensions of a tensor,
    with the last 3 dimensions making up the actual images.

    Process:
    1. check input dtype
    2. move axis
    3. normalize
    4. clamp values
    5. auto scale and convert
    6. convert to rgb
    7. check output dtype

    example:
        Convert a tensor of non-normalised images (..., C, H, W) to a
        tensor of normalised and clipped images (..., H, W, C).
        - integer dtypes are expected to be in the range [0, 255]
        - float dtypes are expected to be in the range [0, 1]

    # TODO: add support for uneven in/out dims, eg. in_dims="HW", out_dims="HWC"
    """
    # 0.a. check tensor
    if not isinstance(tensor, torch.Tensor):
        raise TypeError(f"images must be of type: {torch.Tensor}, got: {type(tensor)}")
    # 0.b. get arguments
    transpose_indices, in_dtype, out_dtype, out_c_dim = _torch_to_images_normalise_args(
        in_tensor_shape=tuple(tensor.shape),
        in_tensor_dtype=tensor.dtype,
        in_dims=in_dims,
        out_dims=out_dims,
        in_dtype=in_dtype,
        out_dtype=out_dtype,
    )
    # 1. check input dtype
    if in_dtype != tensor.dtype:
        raise TypeError(f"images dtype: {repr(tensor.dtype)} does not match in_dtype: {repr(in_dtype)}")
    # 2. move axes
    tensor = tensor.permute(*(i - tensor.ndim for i in range(tensor.ndim - 3)), *transpose_indices)
    # 3. normalise
    if (in_min is not None) or (in_max is not None):
        norm_dtype = out_dtype if out_dtype.is_floating_point else torch.float32
        tensor = torch_image_normalize_channels(
            tensor, in_min=in_min, in_max=in_max, channel_dim=out_c_dim, out_dtype=norm_dtype
        )
    # 4. clamp
    tensor = torch_image_clamp(tensor, clamp_mode=clamp_mode)
    # 5. auto scale and convert
    tensor = torch_image_to_dtype(tensor, out_dtype=out_dtype)
    # 6. convert to rgb
    if always_rgb:
        if tensor.shape[out_c_dim] == 1:
            tensor = torch.repeat_interleave(
                tensor, 3, dim=out_c_dim
            )  # torch.repeat is like np.tile, torch.repeat_interleave is like np.repeat
    # 7. check output dtype
    if out_dtype != tensor.dtype:
        raise RuntimeError(
            f"[THIS IS A BUG!]: After conversion, images tensor dtype: {repr(tensor.dtype)} does not match out_dtype: {repr(in_dtype)}"
        )
    if torch.any(torch.isnan(tensor)):
        raise RuntimeError("[THIS IS A BUG!]: After conversion, images contain NaN values!")
    # convert to numpy
    if to_numpy:
        return tensor.detach().cpu().numpy()
    return tensor


def numpy_to_images(
    ndarray: np.ndarray,
    in_dims: str = "HWC",  # we always treat numpy by default as HWC, and torch.Tensor as CHW
    out_dims: str = "HWC",
    in_dtype: Optional[Union[str, np.dtype]] = None,
    out_dtype: Optional[Union[str, np.dtype]] = np.dtype("uint8"),
    clamp_mode: str = "warn",  # clamp, warn, error
    always_rgb: bool = False,
    in_min: Optional[MinMaxHint] = None,
    in_max: Optional[MinMaxHint] = None,
) -> np.ndarray:
    """
    Convert a batch of image-like arrays to images.
    A batch in this case consists of an arbitrary number of dimensions of an array,
    with the last 3 dimensions making up the actual images.
    - See the docs for: torch_to_images(...)
    """
    # convert numpy dtypes to torch
    if in_dtype is not None:
        in_dtype = _NP_TO_TORCH_DTYPE[np.dtype(in_dtype)]
    if out_dtype is not None:
        out_dtype = _NP_TO_TORCH_DTYPE[np.dtype(out_dtype)]
    # convert back
    array = torch_to_images(
        tensor=torch.from_numpy(ndarray),
        in_dims=in_dims,
        out_dims=out_dims,
        in_dtype=in_dtype,
        out_dtype=out_dtype,
        clamp_mode=clamp_mode,
        always_rgb=always_rgb,
        in_min=in_min,
        in_max=in_max,
        to_numpy=True,
    )
    # done!
    return array


def numpy_to_pil_images(
    ndarray: np.ndarray,
    in_dims: str = "HWC",  # we always treat numpy by default as HWC, and torch.Tensor as CHW
    clamp_mode: str = "warn",
    always_rgb: bool = False,
    in_min: Optional[MinMaxHint] = None,
    in_max: Optional[MinMaxHint] = None,
) -> Union[np.ndarray]:
    """
    Convert a numpy array containing images (..., C, H, W) to an array of PIL images (...,)
    """
    imgs = numpy_to_images(
        ndarray=ndarray,
        in_dims=in_dims,
        out_dims="HWC",
        in_dtype=None,
        out_dtype="uint8",
        clamp_mode=clamp_mode,
        always_rgb=always_rgb,
        in_min=in_min,
        in_max=in_max,
    )
    # all the cases (even ndim == 3)... bravo numpy, bravo!
    images = [Image.fromarray(imgs[idx]) for idx in np.ndindex(imgs.shape[:-3])]
    images = np.array(images, dtype=object).reshape(imgs.shape[:-3])
    # done
    return images


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import lightning as L
import torch

# ========================================================================= #
# Base Modules                                                              #
# ========================================================================= #


class DisentModule(torch.nn.Module):
    def _forward_unimplemented(self, *args):
        # Annoying fix applied by torch for Module.forward:
        # https://github.com/python/mypy/issues/8795
        raise RuntimeError("This should never run!")

    def forward(self, *args, **kwargs):
        raise NotImplementedError


class DisentLightningModule(L.LightningModule):
    # make sure we don't get complaints about the missing methods!
    # -- we prefer to use LightningDataModule
    train_dataloader = None
    test_dataloader = None
    val_dataloader = None
    predict_dataloader = None


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import fields
from numbers import Number
from pprint import pformat
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union
from typing import final

import torch

from disent import registry
from disent.nn.modules import DisentLightningModule
from disent.schedule import Schedule
from disent.util.imports import import_obj

log = logging.getLogger(__name__)


# ========================================================================= #
# framework config                                                          #
# ========================================================================= #


class DisentConfigurable(object):
    @dataclass
    class cfg(object):
        def get_keys(self) -> list:
            return list(self.to_dict().keys())

        def to_dict(self) -> dict:
            return asdict(self)

        def __str__(self):
            return pformat(self.to_dict(), sort_dicts=False)

    def __init__(self, cfg: cfg = cfg()):
        if cfg is None:
            cfg = self.__class__.cfg()
            log.info(f"Initialised default config {cfg=} for {self.__class__.__name__}")
        super().__init__()
        assert isinstance(cfg, self.__class__.cfg), f"{cfg=} ({type(cfg)}) is not an instance of {self.__class__.cfg}"
        self.cfg = cfg


# ========================================================================= #
# framework                                                                 #
# ========================================================================= #


class DisentFramework(DisentConfigurable, DisentLightningModule):
    @dataclass
    class cfg(DisentConfigurable.cfg):
        # optimizer config
        optimizer: Union[
            str
        ] = "adam"  # name in the registry, eg. `adam` OR the path to an optimizer eg. `torch.optim.Adam`
        optimizer_kwargs: Optional[Dict[str, Union[str, float, int]]] = None

    def __init__(
        self,
        cfg: cfg = None,
        # apply the batch augmentations on the GPU instead
        batch_augment: callable = None,
    ):
        # save the config values to the class
        super().__init__(cfg=cfg)
        # check the optimizer
        self.cfg.optimizer = self._check_optimizer(self.cfg.optimizer)
        self.cfg.optimizer_kwargs = self._check_optimizer_kwargs(self.cfg.optimizer_kwargs)
        # batch augmentations may not be implemented as dataset
        # transforms so we can apply these on the GPU instead
        assert callable(batch_augment) or (
            batch_augment is None
        ), f"invalid batch_augment: {repr(batch_augment)}, must be callable or `None`"
        self._batch_augment = batch_augment
        # schedules
        # - maybe add support for schedules in the config?
        self._registered_schedules = set()
        self._active_schedules: Dict[str, Tuple[Any, Schedule]] = {}

    @staticmethod
    def _check_optimizer(optimizer: str):
        if not isinstance(optimizer, str):
            raise TypeError(f"invalid optimizer: {repr(optimizer)}, must be a `str`")
        # check that the optimizer has been registered
        # otherwise check that the optimizer class can be imported instead
        if optimizer not in registry.OPTIMIZERS:
            try:
                import_obj(optimizer)
            except ImportError:
                raise KeyError(
                    f"invalid optimizer: {repr(optimizer)}, valid optimizers are: {sorted(registry.OPTIMIZERS)}, or an import path to an optimizer, eg. `torch.optim.Adam`"
                )
        # return the updated values!
        return optimizer

    @staticmethod
    def _check_optimizer_kwargs(optimizer_kwargs: Optional[dict]):
        # check the optimizer kwargs
        assert isinstance(optimizer_kwargs, dict) or (
            optimizer_kwargs is None
        ), f"invalid optimizer_kwargs type, got: {type(optimizer_kwargs)}"
        # get default kwargs OR copy
        optimizer_kwargs = dict() if (optimizer_kwargs is None) else dict(optimizer_kwargs)
        # set default values
        if "lr" not in optimizer_kwargs:
            optimizer_kwargs["lr"] = 1e-3
            log.info("lr not specified in `optimizer_kwargs`, setting to default value of `1e-3`")
        # return the updated values
        return optimizer_kwargs

    @final
    def configure_optimizers(self):
        # get the optimizer
        # 1. first check if the name has been registered
        # 2. then check if the name can be imported
        if self.cfg.optimizer in registry.OPTIMIZERS:
            optimizer_cls = registry.OPTIMIZERS[self.cfg.optimizer]
        else:
            optimizer_cls = import_obj(self.cfg.optimizer)
        # check that we can call the optimizer
        if not callable(optimizer_cls):
            raise TypeError(f"unsupported optimizer type: {type(optimizer_cls)}")
        # instantiate class
        optimizer_instance = optimizer_cls(self.parameters(), **self.cfg.optimizer_kwargs)
        # check instance
        if not isinstance(optimizer_instance, torch.optim.Optimizer):
            raise TypeError(
                f"returned object is not an instance of torch.optim.Optimizer, got: {type(optimizer_instance)}"
            )
        # return the optimizer
        return optimizer_instance

    @final
    def _compute_loss_step(self, batch, batch_idx, update_schedules: bool):
        try:
            # augment batch with GPU support
            if self._batch_augment is not None:
                batch = self._batch_augment(batch)
            # update the config values based on registered schedules
            if update_schedules:
                # TODO: how do we handle this in the case of the validation and test step? I think this
                #       might still give the wrong results as this is based on the trainer.global_step which
                #       may be incremented by these steps.
                self._update_config_from_schedules()
            # compute loss
            # TODO: move logging into child frameworks?
            loss = self.do_training_step(batch, batch_idx)
            # check loss values
            self._assert_valid_loss(loss)
            self.log("loss", float(loss), prog_bar=True)
            # return loss
            return loss
        except Exception as e:  # pragma: no cover
            # call in all the child processes for the best chance of clearing this...
            # remove callbacks from trainer so we aren't stuck running forever!
            # TODO: this is a hack... there must be a better way to do this... could it be a pl bug?
            #       this logic is duplicated in the run_utils
            if self.trainer and self.trainer.callbacks:
                self.trainer.callbacks.clear()
            # continue propagating errors
            raise e

    @final
    def training_step(self, batch, batch_idx):
        """This is a pytorch-lightning function that should return the computed loss"""
        return self._compute_loss_step(batch, batch_idx, update_schedules=True)

    def validation_step(self, batch, batch_idx):
        """
        TODO: how do we handle the schedule in this case?
        """
        return self._compute_loss_step(batch, batch_idx, update_schedules=False)

    def test_step(self, batch, batch_idx):
        """
        TODO: how do we handle the schedule in this case?
        """
        return self._compute_loss_step(batch, batch_idx, update_schedules=False)

    @final
    def _assert_valid_loss(self, loss):
        if torch.isnan(loss) or torch.isinf(loss):
            raise ValueError("The returned loss is nan or inf")
        if loss > 1e20:
            raise ValueError(f"The returned loss: {loss:.2e} is out of bounds: > {1e+20:.0e}")

    def forward(self, batch) -> torch.Tensor:  # pragma: no cover
        """this function should return the single final output of the model, including the final activation"""
        raise NotImplementedError

    def do_training_step(self, batch, batch_idx) -> torch.Tensor:  # pragma: no cover
        """
        should return a dictionary of items to log with the key 'train_loss'
        as the variable to minimize
        """
        raise NotImplementedError

    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #
    # Schedules                                                             #
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - #

    @final
    def has_schedule(self, target: str):
        return target in self._registered_schedules

    @final
    def remove_schedule(self, target):
        if self.has_schedule(target):
            self._registered_schedules.remove(target)
            self._active_schedules.pop(target)
        else:
            raise KeyError(f"Cannot remove schedule for target {repr(target)} that does not exist!")

    @final
    def register_schedule(self, target: str, schedule: Schedule, logging=True) -> bool:
        """
        returns True if schedule has been activated, False if the key is
        not in the config and it has not be activated!
        """
        assert isinstance(target, str)
        assert isinstance(schedule, Schedule)
        assert type(schedule) is not Schedule
        # handle the case where a schedule for the target already exists!
        if self.has_schedule(target):
            raise KeyError(f"A schedule for target {repr(target)} has already been registered!")
        # check the target exists!
        possible_targets = {f.name for f in fields(self.cfg)}
        # register this target
        self._registered_schedules.add(target)
        # activate this schedule
        if target in possible_targets:
            initial_val = getattr(self.cfg, target)
            self._active_schedules[target] = (initial_val, schedule)
            if logging:
                log.info(f"Activating schedule for target {repr(target)} on {repr(self.__class__.__name__)}.")
            return True
        else:
            if logging:
                log.warning(
                    f"Unactivated schedule for target {repr(target)} on {repr(self.__class__.__name__)} because the key was not found in the config."
                )
            return False

    @final
    def _update_config_from_schedules(self):
        if not self._active_schedules:
            return
        # log the step value
        self.log("scheduler/step", self.trainer.global_step)
        # update the values
        for target, (initial_val, scheduler) in self._active_schedules.items():
            # get the scheduled value
            new_value = scheduler.compute_value(step=self.trainer.global_step, value=initial_val)
            # update it on the config
            setattr(self.cfg, target, new_value)
            # log that things changed
            self.log(f"scheduled/{target}", new_value)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from dataclasses import dataclass
from typing import Sequence

import torch
from torch.distributions import Normal

from disent.frameworks.helper.util import compute_ave_loss_and_logs
from disent.frameworks.vae._unsupervised__betavae import BetaVae
from disent.nn.functional import torch_cov_matrix

# ========================================================================= #
# Dfc Vae                                                                   #
# ========================================================================= #


class DipVae(BetaVae):
    """
    Disentangled Inferred Prior Variational Auto-Encoder (DIP-VAE)
    https://arxiv.org/pdf/1711.00848.pdf

    Reference implementation is from: https://github.com/AntixK/PyTorch-VAE
    """

    REQUIRED_OBS = 1

    @dataclass
    class cfg(BetaVae.cfg):
        dip_mode: str = "ii"
        dip_beta: float = 1.0
        lambda_d: float = 10.0
        lambda_od: float = 5.0

    def __init__(self, model: "AutoEncoder", cfg: cfg = None, batch_augment=None):
        super().__init__(model=model, cfg=cfg, batch_augment=batch_augment)
        # checks
        assert self.cfg.dip_mode in {
            "i",
            "ii",
        }, f'unsupported dip_mode={repr(self.cfg.dip_mode)} for {self.__class__.__name__}. Must be one of: {{"i", "ii"}}'
        assert self.cfg.dip_beta >= 0, "dip_beta must be >= 0"
        assert self.cfg.lambda_d >= 0, "lambda_d must be >= 0"
        assert self.cfg.lambda_od >= 0, "lambda_od must be >= 0"

    # --------------------------------------------------------------------- #
    # Overrides                                                             #
    # --------------------------------------------------------------------- #

    def compute_ave_reg_loss(self, ds_posterior: Sequence[Normal], ds_prior: Sequence[Normal], zs_sampled):
        # compute kl loss
        kl_reg_loss, logs_kl_reg = super().compute_ave_reg_loss(ds_posterior, ds_prior, zs_sampled)
        # compute dip loss
        dip_reg_loss, logs_dip_reg = compute_ave_loss_and_logs(self._dip_compute_loss, ds_posterior)
        # combine
        combined_loss = kl_reg_loss + dip_reg_loss
        # return logs
        return combined_loss, {
            **logs_kl_reg,  # kl_loss, kl_reg_loss
            **logs_dip_reg,  # dip_loss_d, dip_loss_od, dip_loss, dip_reg_loss,
        }

    # --------------------------------------------------------------------- #
    # Helper                                                                #
    # --------------------------------------------------------------------- #

    def _dip_compute_loss(self, d_posterior: Normal):
        cov_matrix = self._dip_estimate_cov_matrix(d_posterior)
        return self._dip_compute_regulariser(cov_matrix)

    def _dip_compute_regulariser(self, cov_matrix):
        """
        Compute DIP regularises for diagonal & off diagonal components
        - covariance matrix should match the identity matrix
        - diagonal and off diagonals are weighted differently
        """
        # covariance diagonal & off diagonal components
        cov_diag = torch.diagonal(cov_matrix)
        cov_off_diag = cov_matrix - torch.diag(cov_diag)
        # regulariser, encourage covariance to match identity matrix
        # TODO: this is all summed, we should rather calculate the means of each so that the scale stays the same after changing z_size
        dip_loss_d = torch.sum((cov_diag - 1) ** 2)  # / num_diag
        dip_loss_od = torch.sum(cov_off_diag**2)  # / num_off_diag
        dip_loss = (self.cfg.lambda_d * dip_loss_d) + (self.cfg.lambda_od * dip_loss_od)
        # scale dip loss - like beta for beta vae
        dip_reg_loss = self.cfg.dip_beta * dip_loss
        # return logs
        return dip_reg_loss, {
            "dip_loss_d": dip_loss_d,
            "dip_loss_od": dip_loss_od,
            "dip_loss": dip_loss,
            "dip_reg_loss": dip_reg_loss,
        }

    def _dip_estimate_cov_matrix(self, d_posterior: Normal):
        z_mean, z_var = d_posterior.mean, d_posterior.variance
        # compute covariance over batch
        cov_z_mean = torch_cov_matrix(z_mean)
        # compute covariance matrix based on mode
        if self.cfg.dip_mode == "i":
            cov_matrix = cov_z_mean
        elif self.cfg.dip_mode == "ii":
            # E[var]
            E_var = torch.mean(torch.diag(z_var), dim=0)
            cov_matrix = cov_z_mean + E_var
        else:  # pragma: no cover
            raise KeyError(f"Unknown DipVAE mode: {self.cfg.dip_mode}")
        # shape: (Z, Z)
        return cov_matrix


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Tuple

import numpy as np

from disent.dataset.data._groundtruth__xysquares import XYSquaresData

# ========================================================================= #
# xy multi grid data                                                        #
# ========================================================================= #


class XColumnsData(XYSquaresData):

    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    Like XYSquares, but has a single column that moves left and right, instead of across a grid.
    - This dataset is also adversarial!
    """

    name = "x_columns"

    @property
    def factor_names(self) -> Tuple[str, ...]:
        return ("x_R", "x_G", "x_B")[: self._num_squares]

    @property
    def factor_sizes(self) -> Tuple[int, ...]:
        return (self._placements,) * self._num_squares

    def _get_observation(self, idx):
        # get factors
        factors = self.idx_to_pos(idx)
        offset, space, size = self._offset, self._spacing, self._square_size
        # GENERATE
        obs = np.zeros(self.img_shape, dtype=self._dtype)
        for i, fx in enumerate(factors):
            x = offset + space * fx
            if self._rgb:
                obs[:, x : x + size, i] = self._fill_value
            else:
                obs[:, x : x + size, :] = self._fill_value
        return obs


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from typing import Callable
from typing import Protocol
from typing import Tuple

import numpy as np
import torch

log = logging.getLogger(__name__)


# ========================================================================= #
# Mining Modes                                                              #
# ========================================================================= #


def _delta_mine_none(dist_ap: torch.Tensor, dist_an: torch.Tensor, top_k: int, margin_max: float):
    assert len(dist_ap) == len(dist_an)
    return torch.arange(len(dist_ap))


def _delta_mine_semi_hard_neg(dist_ap: torch.Tensor, dist_an: torch.Tensor, top_k: int, margin_max: float):
    # SEMI HARD NEGATIVE MINING
    # "choose an anchor-negative pair that is farther than the anchor-positive pair, but within the margin, and so still contributes a positive loss"
    # -- triples satisfy d(a, p) < d(a, n) < alpha
    semi_hard_mask = (dist_ap < dist_an) & (dist_an < margin_max)
    semi_hard_idxs = torch.arange(len(semi_hard_mask))[semi_hard_mask]
    return semi_hard_idxs


def _delta_mine_hard_neg(dist_ap: torch.Tensor, dist_an: torch.Tensor, top_k: int, margin_max: float):
    # HARD NEGATIVE MINING
    # "most similar images which have a different label from the anchor image"
    # -- triples with smallest d(a, n)
    hard_idxs = torch.argsort(dist_an, descending=False)[: int(top_k)]
    return hard_idxs


def _delta_easy_neg(dist_ap, dist_an, cfg):
    # EASY NEGATIVE MINING
    # "least similar images which have the different label from the anchor image"
    raise RuntimeError("This triplet mode is not useful! Choose another.")


def _delta_mine_hard_pos(dist_ap: torch.Tensor, dist_an: torch.Tensor, top_k: int, margin_max: float):
    # HARD POSITIVE MINING -- this performs really well!
    # "least similar images which have the same label to as anchor image"
    # -- shown not to be suitable for all datasets
    hard_idxs = torch.argsort(dist_ap, descending=True)[: int(top_k)]
    return hard_idxs


def _delta_mine_easy_pos(dist_ap: torch.Tensor, dist_an: torch.Tensor, top_k: int, margin_max: float):
    # EASY POSITIVE MINING
    # "the most similar images that have the same label as the anchor image"
    easy_idxs = torch.argsort(dist_ap, descending=False)[: int(top_k)]
    return easy_idxs


_TRIPLET_MINE_MODES = {
    "none": _delta_mine_none,
    "semi_hard_neg": _delta_mine_semi_hard_neg,
    "hard_neg": _delta_mine_hard_neg,
    # 'easy_neg':    delta_mine_easy_neg,  # not actually useful
    "hard_pos": _delta_mine_hard_pos,
    "easy_pos": _delta_mine_easy_pos,
}


# ========================================================================= #
# General Miner                                                             #
# ========================================================================= #


@torch.no_grad()
def mine(mode: str, dist_ap: torch.Tensor, dist_an: torch.Tensor, top_k: int, margin_max: float) -> torch.Tensor:
    # check arrays
    assert (dist_ap.ndim == 1) and (
        dist_an.ndim == 1
    ), f"dist arrays must only have one dimension: dist_ap: {dist_ap.shape} & dist_an: {dist_an.shape}"
    assert dist_ap.shape == dist_an.shape, f"dist array shapes do not match: {dist_ap.shape} & dist_an: {dist_an.shape}"
    # get mining function
    try:
        mine_fn = _TRIPLET_MINE_MODES[mode]
    except KeyError:
        raise KeyError(
            f"invalid triplet mining mode: {repr(mode)}, must be one of: {sorted(_TRIPLET_MINE_MODES.keys())}"
        )
    # mine indices
    idxs = mine_fn(dist_ap=dist_ap, dist_an=dist_an, top_k=top_k, margin_max=margin_max)
    # check and return values
    if len(idxs) > 0:
        return idxs
    else:
        log.warning("no results using {repr(mode)} mining! using entire batch instead")
        return _delta_mine_none(dist_ap=dist_ap, dist_an=dist_an, top_k=top_k, margin_max=margin_max)


def mine_random_mode(mode: str, dist_ap: torch.Tensor, dist_an: torch.Tensor, top_k: int, margin_max: float):
    # randomly choose a mode
    # eg. `ran:hard_pos+easy_pos` randomly chooses between `hard_pos` and `easy_pos`
    if mode.startswith("ran:"):
        mode = np.random.choice(mode[len("ran:") :].split("+"))
    # mine like usual
    return mine(mode=mode, dist_ap=dist_ap, dist_an=dist_an, top_k=top_k, margin_max=margin_max)


# ========================================================================= #
# Helper                                                                    #
# ========================================================================= #


class SampledTripletMineCfgProto(Protocol):
    triplet_margin_max: float
    overlap_num: int
    overlap_mine_triplet_mode: str
    overlap_mine_ratio: float


def configured_mine(dist_ap: torch.Tensor, dist_an: torch.Tensor, cfg: SampledTripletMineCfgProto) -> torch.Tensor:
    return mine_random_mode(
        mode=cfg.overlap_mine_triplet_mode,
        dist_ap=dist_ap,
        dist_an=dist_an,
        top_k=int(cfg.overlap_num * cfg.overlap_mine_ratio),
        margin_max=cfg.triplet_margin_max,
    )


@torch.no_grad()
def configured_idx_mine(
    x_targ: torch.Tensor,
    a_idxs: torch.Tensor,
    p_idxs: torch.Tensor,
    n_idxs: torch.Tensor,
    cfg: SampledTripletMineCfgProto,
    pairwise_loss_fn: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],  # should return arrays with ndim == 1
) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    # TODO: SIMPLIFY THIS FUNCTION HIERARCHY, THERE ARE A LOT OF UNNECESSARY CALLS!
    # TODO: this function is quite useless, its easier to just use configured_mine_random_mode
    # skip mining if mode is None!
    if cfg.overlap_mine_triplet_mode == "none":
        return a_idxs, p_idxs, n_idxs
    # compute differences
    # TODO: this is computationally expensive! sometimes the dist_ap and dist_an may not be used depending on the mode!
    dist_ap = pairwise_loss_fn(x_targ[a_idxs], x_targ[p_idxs])
    dist_an = pairwise_loss_fn(x_targ[a_idxs], x_targ[n_idxs])
    # mine indices
    idxs = configured_mine(dist_ap=dist_ap, dist_an=dist_an, cfg=cfg)
    # check & return values
    return a_idxs[idxs], p_idxs[idxs], n_idxs[idxs]


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import warnings
from typing import Iterable
from typing import Optional
from typing import Sequence

from lightning.pytorch.loggers import Logger
from lightning.pytorch.loggers import WandbLogger

log = logging.getLogger(__name__)

# TODO: convert these functions into generalised log_img, log_vid,
#       log_number, log_text functions that support the different backends.
#       - wandb
#       - disk backend
#       - comet ml

# ========================================================================= #
# Logger Utils                                                              #
# ========================================================================= #


def log_metrics(loggers: Optional[Sequence[Logger]], metrics_dct: dict):
    """
    Log the given values to the given logger.
    - warn the user if something goes wrong
    """
    if loggers:
        for logger in loggers:
            try:
                logger.log_metrics(metrics_dct)
            except:
                warnings.warn(f"Failed to log metrics: {repr(metrics_dct)}")
    else:
        warnings.warn("no trainer.loggers found!")


# ========================================================================= #
# W&B Logger Utils                                                          #
# ========================================================================= #


def wb_yield_loggers(loggers: Optional[Sequence[Logger]]) -> Iterable[WandbLogger]:
    """
    Recursively yield all the loggers or sub-loggers that are an instance of WandbLogger
    """
    if loggers:
        for logger in loggers:
            if isinstance(logger, WandbLogger):
                yield logger


def wb_has_logger(loggers: Optional[Sequence[Logger]]) -> bool:
    for l in wb_yield_loggers(loggers):
        return True
    return False


def wb_log_metrics(loggers: Optional[Sequence[Logger]], metrics_dct: dict):
    """
    Log the given values only to loggers that are an instance of WandbLogger
    """
    wb_logger = None
    # iterate over loggers & update metrics
    for wb_logger in wb_yield_loggers(loggers):
        wb_logger.log_metrics(metrics_dct)
    # warn if nothing logged
    if wb_logger is None:
        warnings.warn("no wandb logger found to log metrics to!")


_SUMMARY_REDICTIONS = {
    "min": min,
    "max": max,
    "recent": lambda prev, current: current,
}


def wb_log_reduced_summaries(loggers: Optional[Sequence[Logger]], summary_dct: dict, reduction="max"):
    """
    Aggregate the given values only to loggers that are an instance of WandbLogger
    - supported reduction modes are `"max"` and `"min"`
    """
    reduce_fn = _SUMMARY_REDICTIONS[reduction]
    wb_logger = None
    # iterate over loggers & update summaries
    for wb_logger in wb_yield_loggers(loggers):
        for key, val_current in summary_dct.items():
            key = f"{key}.{reduction}"
            try:
                val_prev = wb_logger.experiment.summary.get(key, val_current)
                val_next = reduce_fn(val_prev, val_current)
                wb_logger.experiment.summary[key] = val_next
            except:
                log.error(f"W&B failed to update summary for: {repr(key)}", exc_info=True)
    # warn if nothing logged!
    if wb_logger is None:
        warnings.warn("no wandb logger found to log metrics to!")


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from dataclasses import dataclass
from typing import Optional
from typing import Sequence
from typing import final

import torch
from torch.distributions import Normal

from disent.frameworks.helper.reconstructions import ReconLossHandler
from disent.frameworks.helper.reconstructions import make_reconstruction_loss
from disent.frameworks.vae._supervised__adaneg_tvae import AdaNegTripletVae
from disent.nn.loss.triplet_mining import configured_idx_mine

log = logging.getLogger(__name__)


# ========================================================================= #
# Mixin                                                                     #
# ========================================================================= #


class DataOverlapMixin(object):
    # should be inherited by the config on the child class
    @dataclass
    class cfg:
        # override from AE
        recon_loss: str = "mse"
        # OVERLAP VAE
        overlap_loss: Optional[str] = None  # if None, use the value from recon_loss
        overlap_num: int = 1024
        overlap_mine_ratio: float = 0.1
        overlap_mine_triplet_mode: str = "none"
        # AUGMENT
        overlap_augment_mode: str = "augment"
        overlap_augment: Optional[dict] = None

    # private properties
    # - since this class does not have a constructor, it
    #   provides the `init_data_overlap_mixin` method, which
    #   should be called inside the constructor of the child class
    _augment: callable
    _overlap_handler: ReconLossHandler
    _init: bool

    def init_data_overlap_mixin(self):
        if hasattr(self, "_init"):
            raise RuntimeError(
                f"{DataOverlapMixin.__name__} on {self.__class__.__name__} was initialised more than once!"
            )
        self._init = True
        # set augment and instantiate if needed
        if self.cfg.overlap_augment is not None:
            import hydra

            self._augment = hydra.utils.instantiate(self.cfg.overlap_augment)
            assert (self._augment is None) or callable(
                self._augment
            ), f"augment is not None or callable: {repr(self._augment)}, obtained from `overlap_augment={repr(self.cfg.overlap_augment)}`"
        else:
            self._augment = None
        # get overlap loss
        if self.cfg.overlap_loss is None:
            log.info(
                f"`overlap_loss` not specified for {repr(self.__class__.__name__)}, using `recon_loss` instead: {repr(self.cfg.recon_loss)}"
            )
            overlap_loss = self.cfg.recon_loss
        else:
            overlap_loss = self.cfg.overlap_loss
        # construct the overlap handler
        self._overlap_handler: ReconLossHandler = make_reconstruction_loss(name=overlap_loss, reduction="mean")

    @final
    @property
    def overlap_handler(self) -> ReconLossHandler:
        return self._overlap_handler

    def overlap_swap_triplet_idxs(self, x_targ, a_idxs, p_idxs, n_idxs):
        xs_targ = [x_targ[idxs] for idxs in (a_idxs, p_idxs, n_idxs)]
        # CORE: order the latent variables for triplet
        swap_mask = self.overlap_swap_mask(xs_targ=xs_targ)
        # swap all idxs
        swapped_a_idxs = a_idxs
        swapped_p_idxs = torch.where(swap_mask, n_idxs, p_idxs)
        swapped_n_idxs = torch.where(swap_mask, p_idxs, n_idxs)
        # return values
        return swapped_a_idxs, swapped_p_idxs, swapped_n_idxs

    def overlap_swap_mask(self, xs_targ: Sequence[torch.Tensor]) -> torch.Tensor:
        # get variables
        a_x_targ_OLD, p_x_targ_OLD, n_x_targ_OLD = xs_targ
        # CORE OF THIS APPROACH
        # ++++++++++++++++++++++++++++++++++++++++++ #
        # calculate which are wrong!
        # TODO: add more loss functions, like perceptual & others
        with torch.no_grad():
            a_p_losses = self.overlap_handler.compute_pairwise_loss(a_x_targ_OLD, p_x_targ_OLD)  # (B, C, H, W) -> (B,)
            a_n_losses = self.overlap_handler.compute_pairwise_loss(a_x_targ_OLD, n_x_targ_OLD)  # (B, C, H, W) -> (B,)
            swap_mask = a_p_losses > a_n_losses  # (B,)
        # ++++++++++++++++++++++++++++++++++++++++++ #
        return swap_mask

    @torch.no_grad()
    def augment_batch(self, x_targ):
        # ++++++++++++++++++++++++++++++++++++++++++ #
        # perform the augments
        if self.cfg.overlap_augment_mode in ("augment", "augment_each"):
            # recreate augment each time
            if self.cfg.overlap_augment_mode == "augment_each":
                import hydra

                self._augment = hydra.utils.instantiate(self.cfg.overlap_augment)
            # augment on correct device, but skip if not defined!
            aug_x_targ = x_targ if (self._augment is None) else self._augment(x_targ)
        elif self.cfg.overlap_augment_mode == "none":
            # no augment
            aug_x_targ = x_targ
        else:
            raise KeyError(f"invalid cfg.overlap_augment_mode={repr(self.cfg.overlap_augment_mode)}")
        # ++++++++++++++++++++++++++++++++++++++++++ #
        # checks
        assert x_targ.shape == aug_x_targ.shape
        return aug_x_targ

    def mine_triplets(self, x_targ, a_idxs, p_idxs, n_idxs):
        return configured_idx_mine(
            x_targ=x_targ,
            a_idxs=a_idxs,
            p_idxs=p_idxs,
            n_idxs=n_idxs,
            cfg=self.cfg,
            pairwise_loss_fn=self.overlap_handler.compute_pairwise_loss,
        )

    def random_mined_triplets(self, x_targ_orig: torch.Tensor):
        # ++++++++++++++++++++++++++++++++++++++++++ #
        # 1. augment batch
        aug_x_targ = self.augment_batch(x_targ_orig)
        # 2. generate random triples -- this does not generate unique pairs
        a_idxs, p_idxs, n_idxs = torch.randint(
            len(aug_x_targ), size=(3, min(self.cfg.overlap_num, len(aug_x_targ) ** 3)), device=aug_x_targ.device
        )
        # ++++++++++++++++++++++++++++++++++++++++++ #
        # self.debug(x_targ_orig, x_targ, a_idxs, p_idxs, n_idxs)
        # ++++++++++++++++++++++++++++++++++++++++++ #
        # TODO: this can be merged into a single function -- inefficient currently with deltas computed twice
        # 3. reorder random triples
        a_idxs, p_idxs, n_idxs = self.overlap_swap_triplet_idxs(aug_x_targ, a_idxs, p_idxs, n_idxs)
        # 4. mine random triples
        a_idxs, p_idxs, n_idxs = self.mine_triplets(aug_x_targ, a_idxs, p_idxs, n_idxs)
        # ++++++++++++++++++++++++++++++++++++++++++ #
        return a_idxs, p_idxs, n_idxs

    # def debug(self, x_targ_orig, x_targ, a_idxs, p_idxs, n_idxs):
    #     a_p_overlap_orig = - self.recon_handler.compute_unreduced_loss(x_targ_orig[a_idxs], x_targ_orig[p_idxs]).mean(dim=(-3, -2, -1))  # (B, C, H, W) -> (B,)
    #     a_n_overlap_orig = - self.recon_handler.compute_unreduced_loss(x_targ_orig[a_idxs], x_targ_orig[n_idxs]).mean(dim=(-3, -2, -1))  # (B, C, H, W) -> (B,)
    #     a_p_overlap = - self.recon_handler.compute_unreduced_loss(x_targ[a_idxs], x_targ[p_idxs]).mean(dim=(-3, -2, -1))  # (B, C, H, W) -> (B,)
    #     a_n_overlap = - self.recon_handler.compute_unreduced_loss(x_targ[a_idxs], x_targ[n_idxs]).mean(dim=(-3, -2, -1))  # (B, C, H, W) -> (B,)
    #     a_p_overlap_mul = - (a_p_overlap_orig * a_p_overlap)
    #     a_n_overlap_mul = - (a_n_overlap_orig * a_n_overlap)
    #     # check number of things
    #     (up_values_orig, up_counts_orig) = torch.unique(a_p_overlap_orig, sorted=True, return_inverse=False, return_counts=True)
    #     (un_values_orig, un_counts_orig) = torch.unique(a_n_overlap_orig, sorted=True, return_inverse=False, return_counts=True)
    #     (up_values, up_counts) = torch.unique(a_p_overlap, sorted=True, return_inverse=False, return_counts=True)
    #     (un_values, un_counts) = torch.unique(a_n_overlap, sorted=True, return_inverse=False, return_counts=True)
    #     (up_values_mul, up_counts_mul) = torch.unique(a_p_overlap_mul, sorted=True, return_inverse=False, return_counts=True)
    #     (un_values_mul, un_counts_mul) = torch.unique(a_n_overlap_mul, sorted=True, return_inverse=False, return_counts=True)
    #     # plot!
    #     plt.scatter(up_values_orig.detach().cpu(), torch.cumsum(up_counts_orig, dim=-1).detach().cpu())
    #     plt.scatter(un_values_orig.detach().cpu(), torch.cumsum(un_counts_orig, dim=-1).detach().cpu())
    #     plt.scatter(up_values.detach().cpu(), torch.cumsum(up_counts, dim=-1).detach().cpu())
    #     plt.scatter(un_values.detach().cpu(), torch.cumsum(un_counts, dim=-1).detach().cpu())
    #     plt.scatter(up_values_mul.detach().cpu(), torch.cumsum(up_counts_mul, dim=-1).detach().cpu())
    #     plt.scatter(un_values_mul.detach().cpu(), torch.cumsum(un_counts_mul, dim=-1).detach().cpu())
    #     plt.show()
    #     time.sleep(10)


# ========================================================================= #
# Data Overlap Triplet VAE                                                  #
# ========================================================================= #


class DataOverlapTripletVae(AdaNegTripletVae, DataOverlapMixin):
    """
    Michlo et al.
    https://github.com/nmichlo/msc-research

    This is the unsupervised version of the Adaptive Triplet Loss for Variational Auto-Encoders
    - Triplets are sampled from training minibatches and are ordered using some distance
      function taken directly over the datapoints.
    - Adaptive Triplet Loss is used to guide distance learning and encourage disentanglement
    """

    REQUIRED_OBS = 1

    @dataclass
    class cfg(AdaNegTripletVae.cfg, DataOverlapMixin.cfg):
        pass

    def __init__(self, model: "AutoEncoder", cfg: cfg = None, batch_augment=None):
        super().__init__(model=model, cfg=cfg, batch_augment=batch_augment)
        # initialise mixin
        self.init_data_overlap_mixin()

    def hook_compute_ave_aug_loss(
        self, ds_posterior: Sequence[Normal], ds_prior, zs_sampled, xs_partial_recon, xs_targ: Sequence[torch.Tensor]
    ):
        [d_posterior], [x_targ_orig] = ds_posterior, xs_targ
        # 1. randomly generate and mine triplets using augmented versions of the inputs
        a_idxs, p_idxs, n_idxs = self.random_mined_triplets(x_targ_orig=x_targ_orig)
        # 2. compute triplet loss
        loss, loss_log = AdaNegTripletVae.estimate_ada_triplet_loss(
            ds_posterior=[Normal(d_posterior.loc[idxs], d_posterior.scale[idxs]) for idxs in (a_idxs, p_idxs, n_idxs)],
            cfg=self.cfg,
        )
        return loss, {
            **loss_log,
        }


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from typing import Callable

import numpy as np
import torch

from disent.util import to_numpy
from disent.util.visualize import vis_util

log = logging.getLogger(__name__)


# ========================================================================= #
# Visualise Latent Cycle - Modes                                            #
# ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #
# Copyright 2018 The DisentanglementLib Authors. All rights reserved.       #
# Licensed under the Apache License, Version 2.0                            #
# https://github.com/google-research/disentanglement_lib                    #
# Copyright applies to this subsection only.                                #
# ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~ #
# CHANGES:                                                                  #
# - extracted from original code                                            #
# - was not split into functions in this was                                #
# - TODO: convert these functions to torch.Tensors                          #
# ========================================================================= #


def _z_std_gaussian_cycle(base_z, z_means, z_logvars, z_idx, num_frames):
    # Cycle through quantiles of a standard Gaussian.
    zs = np.repeat(np.expand_dims(base_z, 0), num_frames, axis=0)
    zs[:, z_idx] = vis_util.cycle_gaussian(base_z[z_idx], num_frames, loc=0, scale=1)
    return zs


def _z_fitted_gaussian_cycle(base_z, z_means, z_logvars, z_idx, num_frames):
    # Cycle through quantiles of a fitted Gaussian.
    zs = np.repeat(np.expand_dims(base_z, 0), num_frames, axis=0)
    loc = np.mean(z_means[:, z_idx])
    total_variance = np.mean(np.exp(z_logvars[:, z_idx])) + np.var(z_means[:, z_idx])
    zs[:, z_idx] = vis_util.cycle_gaussian(base_z[z_idx], num_frames, loc=loc, scale=np.sqrt(total_variance))
    return zs


def _z_fixed_interval_cycle(base_z, z_means, z_logvars, z_idx, num_frames):
    # Cycle through [-2, 2] interval.
    zs = np.repeat(np.expand_dims(base_z, 0), num_frames, axis=0)
    zs[:, z_idx] = vis_util.cycle_interval(base_z[z_idx], num_frames, -2.0, 2.0)
    return zs


def _z_conf_interval_cycle(base_z, z_means, z_logvars, z_idx, num_frames):
    # Cycle linearly through +-2 std dev of a fitted Gaussian.
    zs = np.repeat(np.expand_dims(base_z, 0), num_frames, axis=0)
    loc = np.mean(z_means[:, z_idx])
    total_variance = np.mean(np.exp(z_logvars[:, z_idx])) + np.var(z_means[:, z_idx])
    scale = np.sqrt(total_variance)
    zs[:, z_idx] = vis_util.cycle_interval(base_z[z_idx], num_frames, loc - 2.0 * scale, loc + 2.0 * scale)
    return zs


def _z_minmax_interval_cycle(base_z, z_means, z_logvars, z_idx, num_frames):
    # Cycle linearly through minmax of a fitted Gaussian.
    zs = np.repeat(np.expand_dims(base_z, 0), num_frames, axis=0)
    zs[:, z_idx] = vis_util.cycle_interval(
        base_z[z_idx], num_frames, np.min(z_means[:, z_idx]), np.max(z_means[:, z_idx])
    )
    return zs


_LATENT_CYCLE_MODES_MAP = {
    "std_gaussian_cycle": _z_std_gaussian_cycle,
    "fitted_gaussian_cycle": _z_fitted_gaussian_cycle,
    "fixed_interval_cycle": _z_fixed_interval_cycle,
    "conf_interval_cycle": _z_conf_interval_cycle,
    "minmax_interval_cycle": _z_minmax_interval_cycle,
}


def make_latent_zs_cycle(
    base_z: torch.Tensor,
    z_means: torch.Tensor,
    z_logvars: torch.Tensor,
    z_idx: int,
    num_frames: int,
    mode: str = "minmax_interval_cycle",
) -> torch.Tensor:
    # get mode
    if mode not in _LATENT_CYCLE_MODES_MAP:
        raise KeyError(f"Unsupported mode: {repr(mode)} not in {set(_LATENT_CYCLE_MODES_MAP)}")
    z_gen_func = _LATENT_CYCLE_MODES_MAP[mode]
    # checks
    assert base_z.ndim == 1
    assert base_z.shape == z_means.shape[1:]
    assert z_means.ndim == z_logvars.ndim == 2
    assert z_means.shape == z_logvars.shape
    assert (
        len(z_means) > 1
    ), f"not enough representations to average, number of z_means should be greater than 1, got: {z_means.shape}"
    # make cycle
    z_cycle = z_gen_func(to_numpy(base_z), to_numpy(z_means), to_numpy(z_logvars), z_idx, num_frames)
    return torch.from_numpy(z_cycle)


# ========================================================================= #
# Visualise Latent Cycles                                                   #
# ========================================================================= #


# TODO: this should be moved into the VAE and AE classes
def make_decoded_latent_cycles(
    decoder_func: Callable[[torch.Tensor], torch.Tensor],
    z_means: torch.Tensor,
    z_logvars: torch.Tensor,
    mode: str = "minmax_interval_cycle",
    num_animations: int = 4,
    num_frames: int = 20,
    decoder_device=None,
) -> torch.Tensor:
    # generate multiple latent traversal visualisations
    animations = []
    for i in range(num_animations):
        frames = []
        for z_idx in range(z_means.shape[1]):
            z = make_latent_zs_cycle(z_means[i], z_means, z_logvars, z_idx, num_frames, mode=mode)
            z = torch.as_tensor(z, device=decoder_device)
            frames.append(decoder_func(z))
        animations.append(torch.stack(frames, dim=0))
    animations = torch.stack(animations, dim=0)
    # return everything
    return animations  # (num_animations, z_size, num_frames, C, H, W)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import os
from abc import ABCMeta
from typing import Callable
from typing import Dict
from typing import NoReturn
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union
from typing import final

import numpy as np

from disent.dataset.util.formats.hdf5 import hdf5_resave_file
from disent.util.function import wrapped_partial
from disent.util.inout.cache import stalefile
from disent.util.inout.files import retrieve_file
from disent.util.inout.paths import filename_from_url
from disent.util.inout.paths import modify_file_name

# ========================================================================= #
# data objects                                                              #
# ========================================================================= #


class DataFile(object, metaclass=ABCMeta):
    """
    base DataFile that does nothing, if the file does
    not exist or it has the incorrect hash, then that's your problem!
    """

    def __init__(self, file_name: str):
        self._file_name = file_name

    @final
    @property
    def out_name(self) -> str:
        return self._file_name

    def prepare(self, out_dir: str) -> str:
        # TODO: maybe check that the file exists or not and raise a FileNotFoundError?
        pass

    def __repr__(self):
        return f"{self.__class__.__name__}(out_name={repr(self.out_name)})"


class DataFileHashed(DataFile, metaclass=ABCMeta):
    """
    Abstract Class
    - Base DataFile class that guarantees a file to exist,
      if the file does not exist, or the hash of the file is
      incorrect, then the file is re-generated.
    """

    def __init__(
        self,
        file_name: str,
        file_hash: Optional[Union[str, Dict[str, str]]],
        hash_type: str = "md5",
        hash_mode: str = "fast",
    ):
        super().__init__(file_name=file_name)
        self._file_hash = file_hash
        self._hash_type = hash_type
        self._hash_mode = hash_mode

    def prepare(self, out_dir: str) -> str:
        @stalefile(
            file=os.path.join(out_dir, self._file_name),
            hash=self._file_hash,
            hash_type=self._hash_type,
            hash_mode=self._hash_mode,
        )
        def wrapped(out_file):
            self._prepare(out_dir=out_dir, out_file=out_file)

        return wrapped()

    def _prepare(self, out_dir: str, out_file: str) -> NoReturn:
        # TODO: maybe raise a FileNotFoundError or a HashError instead?
        raise NotImplementedError


class DataFileHashedDl(DataFileHashed):
    """
    Download a file
    - uri can also be a file to perform a copy instead of download,
      useful for example if you want to retrieve a file from a network drive.
    """

    def __init__(
        self,
        uri: str,
        uri_hash: Optional[Union[str, Dict[str, str]]],
        uri_name: Optional[str] = None,
        hash_type: str = "md5",
        hash_mode: str = "fast",
    ):
        super().__init__(
            file_name=filename_from_url(uri) if (uri_name is None) else uri_name,
            file_hash=uri_hash,
            hash_type=hash_type,
            hash_mode=hash_mode,
        )
        self._uri = uri

    def _prepare(self, out_dir: str, out_file: str):
        retrieve_file(src_uri=self._uri, dst_path=out_file, overwrite_existing=True)

    def __repr__(self):
        return f"{self.__class__.__name__}(uri={repr(self._uri)}, out_name={repr(self.out_name)})"


class DataFileHashedDlGen(DataFileHashed, metaclass=ABCMeta):
    """
    Abstract class
    - download a file and perform some processing on that file.
    """

    def __init__(
        self,
        # download & save files
        uri: str,
        uri_hash: Optional[Union[str, Dict[str, str]]],
        file_hash: Optional[Union[str, Dict[str, str]]],
        # save paths
        uri_name: Optional[str] = None,
        file_name: Optional[str] = None,
        # hash settings
        hash_type: str = "md5",
        hash_mode: str = "fast",
    ):
        self._dl_obj = DataFileHashedDl(
            uri=uri,
            uri_hash=uri_hash,
            uri_name=uri_name,
            hash_type=hash_type,
            hash_mode=hash_mode,
        )
        super().__init__(
            file_name=modify_file_name(self._dl_obj.out_name, prefix="gen") if (file_name is None) else file_name,
            file_hash=file_hash,
            hash_type=hash_type,
            hash_mode=hash_mode,
        )

    def _prepare(self, out_dir: str, out_file: str):
        inp_file = self._dl_obj.prepare(out_dir=out_dir)
        self._generate(inp_file=inp_file, out_file=out_file)

    def _generate(self, inp_file: str, out_file: str):
        raise NotImplementedError

    def __repr__(self):
        return f"{self.__class__.__name__}(uri={repr(self._dl_obj._uri)}, uri_name={repr(self._dl_obj.out_name)}, out_name={repr(self.out_name)})"


class DataFileHashedDlH5(DataFileHashedDlGen):
    """
    Downloads an hdf5 file and pre-processes it into the specified chunk_size.
    """

    def __init__(
        self,
        # download & save files
        uri: str,
        uri_hash: Optional[Union[str, Dict[str, str]]],
        file_hash: Optional[Union[str, Dict[str, str]]],
        # h5 re-save settings
        hdf5_dataset_name: str,
        hdf5_chunk_size: Tuple[int, ...],
        hdf5_compression: Optional[str] = "gzip",
        hdf5_compression_lvl: Optional[int] = 4,
        hdf5_dtype: Optional[Union[np.dtype, str]] = None,
        hdf5_mutator: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        hdf5_obs_shape: Optional[Sequence[int]] = None,
        # save paths
        uri_name: Optional[str] = None,
        file_name: Optional[str] = None,
        # hash settings
        hash_type: str = "md5",
        hash_mode: str = "fast",
    ):
        super().__init__(
            file_name=file_name,
            file_hash=file_hash,
            uri=uri,
            uri_hash=uri_hash,
            uri_name=uri_name,
            hash_type=hash_type,
            hash_mode=hash_mode,
        )
        self._hdf5_resave_file = wrapped_partial(
            hdf5_resave_file,
            dataset_name=hdf5_dataset_name,
            chunk_size=hdf5_chunk_size,
            compression=hdf5_compression,
            compression_lvl=hdf5_compression_lvl,
            out_dtype=hdf5_dtype,
            out_mutator=hdf5_mutator,
            obs_shape=hdf5_obs_shape,
        )
        # save the dataset name
        self._dataset_name = hdf5_dataset_name

    @property
    def dataset_name(self) -> str:
        return self._dataset_name

    def _generate(self, inp_file: str, out_file: str):
        self._hdf5_resave_file(inp_path=inp_file, out_path=out_file)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import numpy as np
from torch import Tensor
from torch import nn

from disent.model import DisentDecoder
from disent.model import DisentEncoder

# ========================================================================= #
# test models                                                               #
# ========================================================================= #


class EncoderLinear(DisentEncoder):
    def __init__(self, x_shape=(3, 64, 64), z_size=6, z_multiplier=1):
        super().__init__(x_shape=x_shape, z_size=z_size, z_multiplier=z_multiplier)

        self.model = nn.Sequential(
            nn.Flatten(), nn.Linear(in_features=int(np.prod(self.x_shape)), out_features=self.z_total)
        )

    def encode(self, x) -> (Tensor, Tensor):
        return self.model(x)


class DecoderLinear(DisentDecoder):
    def __init__(self, x_shape=(3, 64, 64), z_size=6, z_multiplier=1):
        super().__init__(x_shape=x_shape, z_size=z_size, z_multiplier=z_multiplier)

        self.model = nn.Sequential(
            nn.Linear(in_features=self.z_size, out_features=int(np.prod(self.x_shape))),
            nn.Unflatten(dim=1, unflattened_size=self.x_shape),
        )

    def decode(self, z) -> Tensor:
        return self.model(z)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
